"""
This module provides a simple wrapper to enable the instantiation of kafka
producers and consumers
"""

import json
import os
import time
import traceback
import uuid
from ast import literal_eval
from datetime import datetime

import pandas as pd
from confluent_kafka import Producer, Consumer

# TODO: For the kafka consumer implement the commit of the messages only after
#   successful upload to s3. This will reduce the risks of data loss if the pod
#   has errors after reading the data from Kafka
from hip_data_tools.aws.common import AwsConnectionManager, AwsConnectionSettings, AwsSecretsManager
from hip_data_tools.aws.s3 import S3Util
from hip_data_tools.common import get_from_env_or_default_with_warning, LOG

DEFAULT_PRODUCER_CONF = \
    """{'queue.buffering.max.messages': 10000,
    'queue.buffering.max.ms' : 1000}"""
"""Default configurations for the Kafka producer, as string representing a
python dictionary of configurations"""

DEFAULT_CONSUMER_CONF = \
    """{'auto.offset.reset': 'earliest'}"""
"""
Default configurations for the Kafka consumer, as string representing a
python dictionary of configurations
"""

DEFAULT_KAFKA_TIMESTAMP_COLUMN_NAME = 'kafka_timestamp'
"""
Default name of the column in which we store a message's timestamp
provided by Kafka 
"""

DEFAULT_TIMESTAMP_PARTITION_KEY = 'partition_key_ts'
"""
Default column name for the partitioning message batches uploaded to S3
"""


class ProducerConfig:
    """
    Encapsulation of the kafka configurations, reading from environment if
    not supplied, and defaulting where necessary

    Args:
        bootstrap_servers (string): address and port of bootstrap server
        conf (string(dictionary)) : string of dictionary of configurations

    Attributes:
        configs (dict): Dictionary of the kafka configurations

    """

    def __init__(self,
                 bootstrap_servers=None,
                 conf=None):
        self._bootstrap_servers = os.environ.get("KAFKA_BOOTSTRAP_SERVERS") \
            if bootstrap_servers is None else bootstrap_servers

        self.configs = literal_eval(get_from_env_or_default_with_warning(
            "KAFKA_PRODUCER_CONF", DEFAULT_PRODUCER_CONF)) if conf is None \
            else conf

        self.configs.update({'bootstrap.servers': self._bootstrap_servers})


def create_producer(kafka_producer_config=None):
    """
    Creates a Kafka producer

    Args:
       kafka_producer_config (ProducerConfig) : Encapsulated
       configurations for the KafkaProducer
    Returns (Producer): Instantiated Kafka Producer for a specified topic
    """

    kafka_producer_config = kafka_producer_config \
        if kafka_producer_config is not None else ProducerConfig()

    return Producer(kafka_producer_config.configs)


class ConsumerConfig:
    """
    Enscapsulation of the Kafka Consumer Configurations

    Args:
        bootstrap_servers (string): address and port of bootstrap server
        conf (string(dictionary)) : string of dictionary of configurations
        group_id (string): Group Id for the consumer

    Attributes:
        configs (dict): Dictionary of the kafka configurations

    """

    def __init__(self,
                 bootstrap_servers=None,
                 conf=None,
                 group_id=None):
        self.configs = \
            literal_eval(get_from_env_or_default_with_warning(
                "KAFKA_CONSUMER_CONF", DEFAULT_CONSUMER_CONF)) \
                if conf is None else conf

        self._bootstrap_servers = os.environ['KAFKA_BOOTSTRAP_SERVERS'] if \
            bootstrap_servers is None else bootstrap_servers

        self.configs.update({'bootstrap.servers': self._bootstrap_servers})

        # To prevent any clashes with other consumers, if no group id is
        # provided we default to a random uuid
        self._group_id = group_id \
            if group_id is not None else \
            os.environ.get(
                "KAFKA_CONSUMER_GROUP_ID", "random_id_" + str(uuid.uuid4()))

        self.configs.update({'group.id': self._group_id})


def create_consumer(kafka_consumer_config=None):
    """
    Creates a Kafka consumer

    Args:
        kafka_consumer_config (ConsumerConfig): Instantiated kafka
            configuration object
    Returns: Instantiated Kafka Consumer
    """

    kafka_consumer_config = \
        kafka_consumer_config \
            if kafka_consumer_config is not None else ConsumerConfig()

    return Consumer(kafka_consumer_config.configs)


class KafkaPollerConfig:
    """
    Configuration for the poller class

    Args:
        consumer (Consumer): Consumer configurations object
        topic (str): The topic to poll messages from
        timeout_interval (int): Timeout interval for reading the messages
    """

    def __init__(self,
                 consumer=None,
                 topic=None,
                 timeout_interval=None):
        self.consumer = \
            consumer if consumer is not None else create_consumer()

        self.topic = topic if topic is not None else os.environ['KAFKA_TOPIC']
        self.timeout_interval = timeout_interval if timeout_interval is not \
                                                    None \
            else int(os.environ['KAFKA_TIMEOUT_INTERVAL'])


def create_poller(kafka_poller_conf=None):
    """
    Factory method to create a Kafka Poller and take values from environment
    if they aren't provided

    Args:
        kafka_poller_conf (KafkaPollerConfig): Kafka Consumer
    Returns (KafkaPoller): Instantiated KafkaPoller

    """

    poller_config = kafka_poller_conf if kafka_poller_conf is not None \
        else KafkaPollerConfig()

    kafka_poller = KafkaPoller(poller_config.consumer,
                               poller_config.topic,
                               poller_config.timeout_interval)

    return kafka_poller


class BatchS3UploaderConfig:
    """
    Configurations for the BatchS3Uploader

    Args:
        export_path (str): base path within the s3 bucket to export to
        bucket (str): name of the bucket to export results to
        ts_col_nm (str): name given to the time stamp column name
        partition_key_nm (str): name of the column used to store the temporal
            partitioning keys

    Attributes:
        export_path (str): base path within the s3 bucket to export to
        bucket (str): name of the bucket to export results to
        ts_col_nm (str): name given to the time stamp column name
        partition_key_nm (str): name of the column used to store the temporal
            partitioning keys
    """

    def __init__(self,
                 export_path=None,
                 bucket=None,
                 ts_col_nm=None,
                 partition_key_nm=None):
        self.root_path = export_path if export_path is not None \
            else os.environ["KAFKA_EXPORT_PATH"]

        self.bucket = bucket if bucket is not None \
            else os.environ["KAFKA_EXPORT_BUCKET"]

        self.ts_col_nm = ts_col_nm if ts_col_nm is not None else \
            os.environ["KAFKA_TS_COL_NM"]

        self.partition_key_nm = partition_key_nm \
            if partition_key_nm is not None \
            else os.environ["KAFKA_PARTITION_KEY_NM"]


def create_batch_s3_uploader(batch_s3_uploader_config=None):
    """
    Factory method to generate a kafka batch uploadxr

    Args:
        batch_s3_uploader_config (BatchS3UploaderConfig): Configuration
            object for the s3 uploader
    Returns (KafkaS3BatchExporter): Instantiated Exporter

    """

    batch_s3_uploader_config = batch_s3_uploader_config \
        if batch_s3_uploader_config is not None else BatchS3UploaderConfig()

    aws_region = os.environ.get("AWS_DEFAULT_REGION")

    conn = AwsConnectionManager(
        AwsConnectionSettings(region=aws_region, secrets_manager=AwsSecretsManager(), profile=None))

    s3_client = S3Util(conn, batch_s3_uploader_config.bucket)

    return KafkaS3BatchExporter(batch_s3_uploader_config.root_path,
                                s3_client,
                                batch_s3_uploader_config.ts_col_nm,
                                batch_s3_uploader_config.partition_key_nm)


def create_conduit(kafka_poller=None,
                   kafka_exporter=None,
                   polling_interval=None):
    """
    Factory method to create Kafka S3 Conduit, this polls the Kafka queue
    after a set interval and deposits the results onto s3

    Args:
        kafka_poller (KafkaPoller): Kafka Poller used to poll results from a
            topic at set intervals
        kafka_exporter (KafkaS3BatchExporter): Batch Exporter
        polling_interval (int): Time in seconds between polling events

    Returns (KafkaS3Conduit): Instantiated KafkaS3Conduit

    """

    kafka_poller = kafka_poller if kafka_poller is not None else create_poller()
    kafka_exporter = kafka_exporter if kafka_exporter is not None else \
        create_batch_s3_uploader()

    polling_interval = polling_interval if polling_interval is not None else \
        os.environ.get(
            "KAFKA_POLLING_INTERVAL")

    return KafkaConduit(kafka_poller,
                        kafka_exporter,
                        polling_interval)


def generate_snapshot_file_name_with_timestamp():
    """
    Generates the name of the snapshot folder using both the current time
        and date

    Returns (str): String containing the date and time stamped path

    """
    current_datetime = datetime.today().strftime('%Y%m%d_%H%M%S')

    return "df_dt_of_upload_{}.parquet" \
        .format(current_datetime)


def add_interval_partitioning_column(msgs_df,
                                     time_column,
                                     partitioning_key,
                                     interval):
    """
    Adding column to a dataframe with a timestamp column rounded to a time
    interval.

    Args:
        msgs_df (DataFrame): Pandas dataframe to with at least 1 timestamp
            column
        time_column (str): Name of the timestamp column to be rounded
        partitioning_key (str): Name of the column in which we create the
        rounded timestamps
        interval (int): Time interval (in seconds) to round timestamps to

    Returns (DataFame): Returns a dataframe with an appended column of
    rounded timestamps
    """

    msgs_df[time_column] = pd.to_datetime(msgs_df[time_column], unit='ms')
    rounding_time_frame = '{}S'.format(int(interval))
    msgs_df[partitioning_key] = msgs_df[time_column].dt.ceil(
        rounding_time_frame)

    return msgs_df


def convert_msgs_to_dictionary(list_of_msgs):
    """
    Converts json msgs into dictionaries, catching any badly formatted into
    string into their own list

    Args:
        list_of_msgs (list(Message Object)): list of json strings

    Returns (list(dict), list(dict)): Returns correctly formatted dictionaries
        in one list and any ill formatted strings in a second list

    """
    good_data = []
    bad_data = []
    induce_warning = False
    for msg in list_of_msgs:
        msg_body = msg.value().decode('utf-8')
        msg_timestamp = msg.timestamp()[1]
        try:
            dict_of_msg = json.loads(msg_body)
            dict_of_msg[DEFAULT_KAFKA_TIMESTAMP_COLUMN_NAME] = msg_timestamp
            good_data.append(dict_of_msg)
        except Exception as ex:
            induce_warning = True
            bad_data_dict = {DEFAULT_KAFKA_TIMESTAMP_COLUMN_NAME: msg_timestamp,
                             'msg_body': msg_body,
                             'error': type(ex).__name__,
                             'error_arguments': ex.args,
                             'error_traceback': traceback.format_exc()}

            bad_data.append(bad_data_dict)

    if induce_warning:
        LOG.warning("Json decoding error, check bad bucket")

    return good_data, bad_data


class KafkaProducer:
    """
    This class provide a wrapper around the lower level Kafka produce,
    and extends this base functionality with commonly used methods or
    approaches to dealing with a message queue


    Args:
        topic (str): Name of Kafka topic to publish the message to
        raise_exception_on_failed_connection (bool): raise exception if the
            Kafka broker cannot be polled
        bootstrap_servers:  (str): String list of bootstrap servers and their
            ports e.g. 'localhost:9092' or '[broker1:9092, broker2:9092]'
        config (str):string  representation of a python dictionary which
            encapsulates the required settings for the consumer

    Attributes:
        topic (str): Name of Kafka topic to publish the message to
        bootstrap_servers:  (str): String list of bootstrap servers and their
            ports e.g. 'localhost:9092' or '[broker1:9092, broker2:9092]'
        config (str):string  representation of a python dictionary which
            encapsulates the required settings for the consumer
        producer (Producer): Kafka producer class

    """

    def __init__(self,
                 topic=None,
                 raise_exception_on_failed_connection=False,
                 bootstrap_servers=None,
                 config=None):

        self.topic = os.environ.get("KAFKA_TOPIC") if topic is None else topic
        self.producer_config = ProducerConfig(bootstrap_servers, config)
        self.producer = None
        self.raise_exception_on_failed_connection = \
            raise_exception_on_failed_connection

    def __del__(self):
        """On destruction flush the local queue to ensure all messages are
        actually pushed to the relevant topic

        Returns : None
        """

        if self.producer is not None:
            self.producer.flush(1)

    def instantiate_producer(self):
        """
        Try to connect to the Kafka bootstrap server. We include
        functionality to allow failure to connect to the queue to happen
        silently

        Returns (Producer): Kafka Producer

        """
        try:
            LOG.debug("Instantiating Producer")
            self.producer = create_producer(self.producer_config)
            # Check the connection works by polling for messages
            LOG.debug("Polling Queue")
            self.producer.poll(3)
            LOG.info("Succesfully polled Kafka Queue")

        except Exception as exception:
            self.producer = None
            LOG.error("Kafka Producer failed to instantiate: \n %s", exception)

            if self.raise_exception_on_failed_connection:
                raise NoProducerInstantiatedError()

    def _instantiated_producer_if_required(self):
        if self.producer is None:
            self.instantiate_producer()

    def publish_dict_as_json(self, in_dict):
        """
        Converts a dictionary to json and pushes it to the Kafka topic

        Args:
            in_dict (dict): Dictionary to be pushed to the topic

        Returns (string): Message published to kafka

        """
        json_string = json.dumps(in_dict)
        self.produce_msg(json_string)
        return json_string

    def produce_msg(self, msg):
        """
        Produces a message to the Kafka topic

        Args:
            msg (str): String to be push to the Kafka topic
        Returns: None

        """
        self._instantiated_producer_if_required()

        if self.producer is not None:
            LOG.debug("Producing message on topic %s : %s", self.topic, msg)
            self.producer.produce(self.topic, msg)
        else:
            LOG.warning(
                "Kafka Connection not initialised, message not sent, "
                "message body: %s", msg)

            if self.raise_exception_on_failed_connection:
                raise NoProducerInstantiatedError


class KafkaPoller:
    """
    This class provides a wrapper around the Kafka consumer to enable polling
    of a kafka queue and aggregation of all messages into a single list

    Args:
        kafka_consumer (Consumer): Kafka Consumer
        topic (str): Topic to subscribe to
        timeout_interval (int): Time int seconds before the consumer times out

    Attributes:
        kafka_consumer (Consumer): Kafka Consumer
        topic (str): Topic to subscribe to
        timeout_interval (int): Time int seconds before the consumer times out
    """

    def __init__(self,
                 kafka_consumer,
                 topic,
                 timeout_interval):

        self._kafka_consumer = kafka_consumer
        self._topic = topic
        self._timeout_interval = timeout_interval
        self._subscribed_to_topic = False

    def _subscribe_consumer(self):
        """
        Subscribe the Kafka consumer to the given topic

        Returns: None

        """
        self._kafka_consumer.subscribe([self._topic])
        self._subscribed_to_topic = True

    def get_msgs(self):
        """Get the latest messages from the Kafka topic

        Returns list(Message) : list of Kafka Messages
        """

        LOG.debug("Getting messages from topic %s", self._topic)

        if not self._subscribed_to_topic:
            self._subscribe_consumer()

        return self.poll_kafka_for_messages()

    def poll_kafka_for_messages(self):
        """
        Poll the kafka topic for messages

        Returns list(Message): list of Kafka messages

        """
        list_of_mgs = []
        while True:
            msg = self._kafka_consumer.poll(self._timeout_interval)

            if msg is None:
                break

            msg = self._check_msg_is_not_error(msg)

            if msg is not None:
                list_of_mgs.append(msg)

        return list_of_mgs

    @staticmethod
    def _check_msg_is_not_error(msg):

        if msg.error():
            LOG.error("Consumer error: %s", msg.error())
            return None

        LOG.debug(
            "Message from topic: %s", msg.value().decode('utf-8'))

        return msg


class KafkaS3BatchExporter:
    """
    This class enables the deposition of a batch of Kafka messages into a s3
    bucket at set intervals
    """

    def __init__(self,
                 root_path,
                 s3_client,
                 ts_col_nm,
                 partition_key_nm):

        self.root_path = root_path
        self.ts_col_nm = ts_col_nm
        self.partition_key_nm = partition_key_nm
        self._s3_client = s3_client

    def parse_and_export_msgs(self,
                              list_of_msgs,
                              interval):
        """
        Converts messages to a pandas dataframe and then exports to s3

        Args:
            list_of_msgs (list(Kafka Message Object)): List of msg objects
            interval (int): Rounding interval for the temporal partitioning
        Returns: None

        """

        good_data, bad_data = convert_msgs_to_dictionary(list_of_msgs)

        self._partition_data_and_upload_to_s3(good_data, interval)
        self._partition_data_and_upload_to_s3(bad_data, interval)

        LOG.info("Data Upload Complete")

    def _partition_data_and_upload_to_s3(self, data_list, interval):
        """
        Partitions the messages by time in a dataframe, and then uploads to s3

        """
        if data_list:

            for msg_df, s3_path in \
                self.partition_msgs_by_kafka_ts(data_list, interval):
                LOG.debug("data path : %s", s3_path)

                self._s3_client.upload_dataframe_as_parquet(msg_df, s3_path)

    def partition_msgs_by_kafka_ts(self, list_of_dicts, interval):
        """
        Partition messages by their Kafka timestamps and uses these
        timestamps to generate their relevant paths on s3
        Args:
            list_of_dicts (list(dict)): list of dictionaries to upload to s3
            interval (int): Rounding interval for the temporal partitioning
        Returns list((DataFrame, string)): Returns the dataframe for each
        temporal partition, and the path to upload it to

        """
        df_msgs_and_meta_data = pd.DataFrame(list_of_dicts)

        df_msgs_and_meta_data = \
            add_interval_partitioning_column(df_msgs_and_meta_data,
                                             self.ts_col_nm,
                                             self.partition_key_nm,
                                             int(interval))
        data_frames_and_fld_locs = self.generate_partitioned_dataframes(
            df_msgs_and_meta_data)

        return data_frames_and_fld_locs

    def generate_partitioned_dataframes(self, df_msgs_and_meta_data):
        """
        Generates the batched dataframes to upload to s3
        Args:
            df_msgs_and_meta_data (dataframe): dataframe to be paritioned
        Returns (List(dataframe, string)): list of dataframes and thier
            folder paths on s3
        """
        dataframes_and_fld_locs = []
        for cur_interval_np_datetime in \
            pd.unique(df_msgs_and_meta_data[self.partition_key_nm]):
            df_partition = \
                df_msgs_and_meta_data[
                    df_msgs_and_meta_data[self.partition_key_nm] ==
                    cur_interval_np_datetime]

            cur_interval_ts = pd.to_datetime(cur_interval_np_datetime)

            batch_file_path = "{}/date_of_batch={}/time_of_batch={}".format(
                self.root_path,
                cur_interval_ts.strftime('%Y%m%d'),
                cur_interval_ts.strftime('%H%M%S')
            )

            file_nm = generate_snapshot_file_name_with_timestamp()

            full_path_on_s3 = '{}/{}'.format(batch_file_path, file_nm)

            LOG.debug("Data path : %s", full_path_on_s3)

            dataframes_and_fld_locs.append((df_partition, full_path_on_s3))

        return dataframes_and_fld_locs


class KafkaConduit:
    """
    This class polls the Kafka topic at set intervals and exports the results
    to an s3 bucket

    Args:
        kafka_poller (KafkaPoller): Instantiated Kafka Poller
        kafka_exporter (KafkaS3BatchExporter): Exporter to parse and export
            Kafka messages
        Polling Interval (int): Interval in Seconds at which to poll the
            Kafka topic and export the messages to s3
    """

    def __init__(self,
                 kafka_poller,
                 kafka_exporter,
                 polling_interval):
        self._kafka_poller = kafka_poller
        self._kafka_s3_exporter = kafka_exporter
        self._polling_interval = polling_interval

    def poll_topic_and_upload_to_s3(self):
        """
        Poll at the Kafka topic at set intervals and parse and export the
        messages to S3
        Returns: None
        """
        while True:
            LOG.debug("Polling Kafka for messages")
            self.create_events_snapshot()
            log.debug("Upload complete, sleeping")
            time.sleep(self._polling_interval)

    def create_events_snapshot(self):
        """
        Get Kafka messages from a topic and export to s3

        Returns: None

        """

        msgs = self._kafka_poller.get_msgs()
        log.debug("Json messages : %s", msgs)

        self._kafka_s3_exporter.parse_and_export_msgs(msgs,
                                                      self._polling_interval)


class NoProducerInstantiatedError(Exception):
    """Exception raised when no producer has been instantiated"""



"""
Common ETL specific utilities and methods
"""
import string
from datetime import datetime
from enum import Enum
from random import random

import time
from cassandra.cqlengine import columns, ValidationError
from cassandra.cqlengine.management import sync_table
from cassandra.cqlengine.models import Model
from cassandra.cqlengine.query import LWTException


class EtlStates(Enum):
    """Enumerator for the possible states of an ETL"""
    Ready = 'ready'
    Processing = 'processing'
    Succeeded = 'succeeded'
    Failed = 'failed'


class EtlSinkRecordState(Model):
    """Cassandra ORM model for the Etl Sink States"""
    etl_signature = columns.Text(primary_key=True, required=True)
    record_identifier = columns.Text(primary_key=True, required=True)
    record_state = columns.Text(required=True, default=EtlStates.Ready.value)
    state_created = columns.DateTime(required=True, default=datetime.now())
    state_last_updated = columns.DateTime(required=True, default=datetime.now())


def sync_etl_state_table():
    """
    Utility method to sync (Create) the table as per ORM model
    Returns: None
    """
    sync_table(EtlSinkRecordState)


def _state_change_validation(current: EtlStates, new: EtlStates) -> None:
    if new in (EtlStates.Succeeded, EtlStates.Failed) and current != EtlStates.Processing:
        raise ValidationError(
            f"Current State '{current}' cannot transition to '{new}'")

    if new == EtlStates.Processing and current != EtlStates.Ready:
        raise ValidationError(
            f"Current State '{current}' cannot transition to '{new}'")

    if new == current:
        raise ValidationError(
            f"Current State '{current}' cannot transition to '{new}', "
            f"you might have duplicate records in your data set")

    if current == EtlStates.Succeeded:
        raise ValidationError(
            f"Current State '{current}' does not allow any transitions")


class EtlSinkRecordStateManager:
    """
    The Generic ETL Sink State manager that manages and persists the state of a record
    Args:
        record_identifier (str): A unique Identifier string to identify the sink record
        etl_signature (str): The Unique ETL Signature to identify the ETL
    """

    def __init__(self, record_identifier: str, etl_signature: str):
        self.record_identifier = record_identifier
        self.etl_signature = etl_signature
        self.remote_state = self._get_or_create_state()

    def _get_or_create_state(self):
        try:
            return EtlSinkRecordState.if_not_exists().create(
                etl_signature=self.etl_signature,
                record_identifier=self.record_identifier)
        except LWTException as e:
            return EtlSinkRecordState.get(
                etl_signature=e.existing['etl_signature'],
                record_identifier=e.existing['record_identifier'])

    def _change_state(self, new_state):
        _state_change_validation(self.current_state(), new_state)

        self.remote_state.record_state = new_state.value
        self.remote_state.state_last_updated = datetime.now()
        self.remote_state.save()

    def current_state(self) -> EtlStates:
        """
        Get current state of the sid record
        Returns (EtlStates): current state

        """
        return EtlStates(self.remote_state.record_state)

    def succeeded(self) -> None:
        """
        Mark Record as succeeded
        Returns: None
        """
        self._change_state(EtlStates.Succeeded)

    def failed(self) -> None:
        """
        Mark Record as failed
        Returns: None
        """
        self._change_state(EtlStates.Failed)

    def processing(self) -> None:
        """
        Mark Record as processing
        Returns: None
        """
        self._change_state(EtlStates.Processing)

    def ready(self) -> None:
        """
        Mark Record as ready
        Returns: None
        """
        self._change_state(EtlStates.Ready)


def current_epoch() -> int:
    """
    Get the current epoch to millisecond precision
    Returns: int
    """
    return int(round(time.time() * 1000))


def get_random_string(length: int) -> str:
    """
    A random ascii lowercase string of certain length
    Args:
        length (int): length of the random string
    Returns: str
    """
    letters = string.ascii_lowercase
    return ''.join(random.choice(letters) for i in range(length))



"""
handle ETL of data from Athena to Cassandra
"""
from typing import List

from attr import dataclass
from cassandra.cqlengine import ValidationError
from pandas import DataFrame

from hip_data_tools.apache.cassandra import CassandraConnectionManager, CassandraConnectionSettings
from hip_data_tools.common import LOG
from hip_data_tools.etl.athena_to_dataframe import AthenaToDataFrame, AthenaToDataFrameSettings
from hip_data_tools.etl.common import EtlSinkRecordStateManager, sync_etl_state_table, EtlStates
from hip_data_tools.google.adwords import AdWordsOfflineConversionUtil, \
    GoogleAdWordsConnectionManager, GoogleAdWordsConnectionSettings


@dataclass
class AthenaToAdWordsOfflineConversionSettings(AthenaToDataFrameSettings):
    """S3 to Cassandra ETL settings"""
    transformation_column_mapping: dict
    etl_identifier: str
    etl_state_manager_keyspace: str
    etl_state_manager_connection: CassandraConnectionSettings
    destination_batch_size: int
    destination_connection_settings: GoogleAdWordsConnectionSettings


def _get_record_signature(record: dict):
    return f"{record['googleClickId']}||{record['conversionName']}||{record['conversionTime']}"


def _get_structured_issue(error, data):
    return {
        "error": error,
        "data": data,
    }


class AthenaToAdWordsOfflineConversion(AthenaToDataFrame):
    """
    Class to transfer parquet data from s3 to Cassandra
    Args:
        settings (AthenaToCassandraSettings): the settings around the etl to be executed
    """

    def __init__(self, settings: AthenaToAdWordsOfflineConversionSettings):
        self.__settings = settings
        super().__init__(settings)
        self._adwords = None

    def upload_next(self) -> List[dict]:
        """
        Upload the next file in line from the athena table onto AdWords offline conversion
        Returns List[dict]: a list of issues in the format
        [
            {
                "error": " some error",
                "data": {
                ... original data body of the data that caused issues
                }
            },
        ]
        """
        return self._process_data_frame(self.next())

    def upload_all(self) -> List[dict]:
        """
        Upload all files from the Athena table onto AdWords offline conversion
        Returns List[dict]: a list of issues in the format
        [
            {
                "error": " some error",
                "data": {
                ... original data body of the data that caused issues
                }
            }.
        ]
        """
        issues = []
        for key in self.list_source_files():
            issues.extend(self._process_data_frame(self.get_data_frame(key)))
        return issues

    def _get_adwords_util(self):
        if self._adwords is None:
            self._adwords = AdWordsOfflineConversionUtil(
                GoogleAdWordsConnectionManager(self.__settings.destination_connection_settings)
            )
        return self._adwords

    def _get_sink_manager(self, record: dict) -> EtlSinkRecordStateManager:
        # Need to set up the cassandra connection
        return EtlSinkRecordStateManager(
            record_identifier=_get_record_signature(record),
            etl_signature=self.__settings.etl_identifier
        )

    def _data_frame_to_destination_dict(self, data_frame: DataFrame) -> List[dict]:
        data_frame = data_frame.rename(columns=self.__settings.transformation_column_mapping)
        approved_fields = self._get_adwords_util().valid_fields
        drop_fields = [col for col in list(data_frame.columns) if col not in approved_fields]
        data_frame = data_frame.drop(drop_fields, axis=1)
        return data_frame.to_dict('records')

    def _chunk_batches(self, lst: List[dict]) -> List[List[dict]]:
        n = self.__settings.destination_batch_size
        return [lst[i * n:(i + 1) * n] for i in range((len(lst) + n - 1) // n)]

    def _process_data_frame(self, data_frame) -> List[dict]:
        data_dict = self._data_frame_to_destination_dict(data_frame)
        self._state_manager_connect()
        ready_data, issues = self._verify_data_before_upsert(data_dict)
        data_dict_batches = self._chunk_batches(ready_data)
        for data_batch in data_dict_batches:
            data_to_process, processing_issue = self._mark_processing(data_batch)
            issues.extend(processing_issue)
            success, fail = self._upload_conversions(data_to_process)
            self._mark_upload_results(fail, success)
        return issues

    def _upload_conversions(self, data_batch):
        return self._get_adwords_util().upload_conversions(data_batch)

    def _state_manager_connect(self):
        conn = CassandraConnectionManager(self.__settings.etl_state_manager_connection)
        conn.setup_connection(self.__settings.etl_state_manager_keyspace)
        sync_etl_state_table()

    def _mark_processing(self, data: List[dict]) -> (List[dict], List[dict]):
        data_for_processing = []
        issues = []
        for dat in data:
            try:
                self._get_sink_manager(dat).processing()
                data_for_processing.append(dat)
            except ValidationError as e:
                issues.append(_get_structured_issue(str(e), dat))
        return data_for_processing, issues

    def _mark_upload_results(self, fail: List[dict], success: List[dict]) -> None:
        for dat in success:
            self._get_sink_manager(dat).succeeded()
        for dat in fail:
            self._get_sink_manager(dat["data"]).failed()

    def _verify_data_before_upsert(self, data: List[dict]) -> (List[dict], List[dict]):
        data, issues = map(list, zip(*[self._sanitise_data(dat) for dat in data]))
        # Remove None from the List
        return [i for i in data if i], [i for i in issues if i]

    def _sanitise_data(self, dat):
        try:
            if self._verify_state(dat):
                return dat, None
            else:
                return None, _get_structured_issue(f"Current state is not Ready", dat)
        except ValidationError as e:
            LOG.warning("Issue while trying to ready a record for the upload \n %s \n %s", e,
                        dat)
            return None, _get_structured_issue(str(e), dat)

    def _verify_state(self, data):
        current_state = self._get_sink_manager(data).current_state()
        return current_state == EtlStates.Ready



"""
Module to deal with data transfer from Google sheets to S3
"""

from attr import dataclass

from hip_data_tools.aws.common import AwsConnectionManager
from hip_data_tools.aws.common import AwsConnectionSettings
from hip_data_tools.aws.s3 import S3Util
from hip_data_tools.google.common import GoogleApiConnectionSettings
from hip_data_tools.google.sheets import SheetUtil, GoogleSheetConnectionManager


@dataclass
class GoogleSheetsToS3Settings:
    """
    Google sheets to Athena ETL settings
    Args:
        source_workbook_url: str
        source_sheet: str
        source_row_range: str
        source_field_names_row_number: int
        source_field_types_row_number: int
        source_data_start_row_number: int
        source_connection_settings: GoogleApiConnectionSettings
        manual_partition_key_value: dict
        target_s3_bucket: str
        target_s3_dir: str
        target_connection_settings: AwsConnectionSettings
    """
    source_workbook_url: str
    source_sheet: str
    source_row_range: str
    source_field_names_row_number: int
    source_field_types_row_number: int
    source_data_start_row_number: int
    source_connection_settings: GoogleApiConnectionSettings
    manual_partition_key_value: dict
    target_s3_bucket: str
    target_s3_dir: str
    target_connection_settings: AwsConnectionSettings


class GoogleSheetToS3:
    """
    Class to transfer data from google sheet to athena
    Args:
        settings (GoogleSheetsToAthenaSettings): the settings around the etl to be executed
    """

    def __init__(self, settings: GoogleSheetsToS3Settings):
        self.__settings = settings
        self.data = None

    def _get_sheets_util(self):
        return SheetUtil(
            conn_manager=GoogleSheetConnectionManager(self.__settings.source_connection_settings),
            workbook_url=self.__settings.source_workbook_url,
            sheet=self.__settings.source_sheet)

    def _get_s3_util(self):
        return S3Util(
            bucket=self.__settings.target_s3_bucket,
            conn=AwsConnectionManager(settings=self.__settings.target_connection_settings))

    def _get_sheet_dataframe(self):
        if self.data is None:
            sheet_util = self._get_sheets_util()
            self.data = sheet_util.get_dataframe(
                field_names_row_number=self.__settings.source_field_names_row_number,
                field_types_row_number=self.__settings.source_field_types_row_number,
                row_range=self.__settings.source_row_range,
                data_start_row_number=self.__settings.source_data_start_row_number)
        return self.data

    def write_sheet_data_to_s3(self):
        """
        Write the data frame into S3
        Args:
            s3_key (string): s3 key
        :return: None
        """
        s3_util = self._get_s3_util()
        s3_util.upload_dataframe_as_parquet(
            dataframe=self._get_sheet_dataframe(),
            key=self.__settings.target_s3_dir,
            file_name="sheet_data")



"""
This module is responsible for all Google sheets operations
"""
import re
from typing import List, Any

import pandas as pd
from pandas import DataFrame
import gspread

from hip_data_tools.common import LOG
from hip_data_tools.google.common import GoogleApiConnectionManager, GoogleApiConnectionSettings


class GoogleSheetConnectionManager(GoogleApiConnectionManager):
    """Encapsulates the Google sheets API connection settings"""

    def __init__(self, settings: GoogleApiConnectionSettings):
        super().__init__(settings, scope=['https://spreadsheets.google.com/feeds',
                                          'https://www.googleapis.com/auth/drive'])

    def get_client(self):
        """
        Get the connection for google sheets
        Returns: authorised connection for google sheets
        """
        return gspread.authorize(self._credentials())


def _get_value_matrix_from_range(worksheet, row_range):
    list_of_lists = []
    row_numbers = [int(i) for i in row_range.split(':')]
    first_row_number = row_numbers[0]
    last_row_number = row_numbers[1]
    for i in range(first_row_number, last_row_number + 1):
        row = worksheet.row_values(i)
        list_of_lists.append(row)
    return list_of_lists


def _get_value_matrix_from_skip(worksheet, data_start_row_number):
    list_of_lists = worksheet.get_all_values()
    del list_of_lists[0:data_start_row_number - 1]
    return list_of_lists


def _validate_field_names_and_types_count(field_names, field_types):
    if len(field_names) != len(field_types):
        LOG.error("Number of field names and number of field types are not matching")
        raise Exception("Field names and types are not matching")


def _validate_field_names(field_names):
    for field_name in field_names:
        # check for strings which only contains letters, numbers and underscores
        if not re.match("^[A-Za-z0-9_]+$", field_name):
            LOG.error("Unsupported field name: %s", field_name)
            raise Exception("Unsupported field name")


def _list_of_list_to_list_of_tuples(data: List[List[Any]]) -> List[tuple]:
    return [tuple(l) for l in data]


def _simplified_dtype(data_type):
    """
    Return the athena base data type
    Args:
        data_type (string): data type
    :return: simplified data type
    """
    return ((re.sub(r'\(.*\)', '', data_type)).split(" ", 1)[0]).upper()


def _get_typed_data_frame(data_frame: DataFrame, data_types: List[str],
                          field_names: List[str]) -> DataFrame:
    for data_type, field_name in zip(data_types, field_names):
        if data_type == 'NUMBER':
            data_frame[[field_name]] = data_frame[[field_name]].apply(pd.to_numeric)
        # TODO need to add date type support
    return data_frame


class SheetUtil:
    """
    Utility class for connecting to google sheets
    Args:
        conn_manager (GoogleSheetConnectionManager): connection to read google sheets
        conn_manager (integer): row number of the filed names
        conn_manager (integer): row number of the field types
    """

    def __init__(self, conn_manager: GoogleSheetConnectionManager, workbook_url, sheet):
        self.google_sheet_connection = conn_manager.get_client()
        self.workbook_url = workbook_url
        self.sheet = sheet
        self._worksheet = None

    def _get_worksheet(self):
        if self._worksheet is None:
            self._worksheet = self.google_sheet_connection.open_by_url(self.workbook_url).worksheet(
                self.sheet)
        return self._worksheet

    def get_value_matrix(self, row_range=None, data_start_row_number=None) -> List[List[Any]]:
        """
        Get the values of the sheet as a 2D array
        Args:
            row_range (string): [optional]range of the rows that need to be selected (eg: '1:7')
            data_start_row_number (integer): (eg: 2)
        Returns: values of the sheet as a 2D array
        """
        if not row_range and not data_start_row_number:
            ValueError(
                "Both row_range and data_start_row_number cannot be None, provide at least one")

        if row_range:
            return _get_value_matrix_from_range(worksheet=self._get_worksheet(),
                                                row_range=row_range)

        return _get_value_matrix_from_skip(worksheet=self._get_worksheet(),
                                           data_start_row_number=data_start_row_number)

    def get_fields_names(self, field_names_row_number: int) -> List[str]:
        """
        Get the field names as a list
        Args:
            field_names_row_number (int): row number of field names
        Returns: field names list
        """
        field_names = self._get_worksheet().row_values(field_names_row_number)
        LOG.debug("Field names:\n%s", field_names)
        return field_names

    def get_fields_types(self, field_types_row_number: int) -> List[str]:
        """
        Get the field types as a list
        Args:
            field_types_row_number (int): Row number of field types
        Returns: field types list
        """
        field_types = self._get_worksheet().row_values(field_types_row_number)
        LOG.debug("Field types:\n%s", field_types)
        return field_types

    def get_dataframe(self,
                      field_names_row_number: int,
                      field_types_row_number: int,
                      row_range=None,
                      data_start_row_number=None) -> DataFrame:
        """
        Get the field types as a list
        Args:
            field_names_row_number (int): Row number of field names
            field_types_row_number (int): Row number of field types
            row_range (str): Row range eg: "3:4"
            data_start_row_number (int): Starting row number of actual data
        Returns: field types list
        """
        matrix = self.get_value_matrix(row_range, data_start_row_number)
        return self._get_dataframe_from_matrix(field_names_row_number, field_types_row_number,
                                               matrix)

    def _get_dataframe_from_matrix(self, field_names_row_number, field_types_row_number, matrix):
        tupled_data = _list_of_list_to_list_of_tuples(matrix)
        field_names = self.get_fields_names(field_names_row_number)
        untyped_data_frame = DataFrame(data=tupled_data,
                                       columns=field_names)
        typed_data_frame = _get_typed_data_frame(data_frame=untyped_data_frame,
                                                 data_types=self.get_fields_types(
                                                     field_types_row_number),
                                                 field_names=field_names)
        return typed_data_frame



"""
Utility for connecting to and transforming data in Cassandra clusters
"""
import os
from typing import List

import pandas as pd
from attr import dataclass
from cassandra import ConsistencyLevel
from cassandra.auth import PlainTextAuthProvider
from cassandra.cluster import Cluster, Session
from cassandra.cqlengine import connection
from cassandra.cqlengine.management import sync_table
from cassandra.datastax.graph import Result
from cassandra.policies import LoadBalancingPolicy
from cassandra.query import dict_factory, BatchStatement, PreparedStatement
from pandas import DataFrame
from pandas._libs.tslibs.nattype import NaT
from pandas._libs.tslibs.timestamps import Timestamp
from retrying import retry

from hip_data_tools.common import KeyValueSource, ENVIRONMENT, SecretsManager, LOG

_RETRY_WAIT_MULTIPLIER_MS: int = int(os.getenv("CASSANDRA_RETRY_WAIT_MULTIPLIER_MS", "1000"))
"""Exponential backoff settings for connections to cassandra"""

_RETRY_WAIT_MAX_MS: int = int(os.getenv("CASSANDRA_RETRY_WAIT_MAX_MS", "100000"))
"""Exponential backoff settings for connections to cassandra"""

_PYTHON_TO_CASSANDRA_DATA_TYPE_MAP = {
    "Timestamp": "timestamp",
    "str": "varchar",
    "int64": "bigint",
    "int32": "int",
    "dict": "map",
    "float64": "double",
    "UUID": "UUID",
}
"""Dictionary mapping of python and pandas data types to Cassandra data types"""


def _get_data_frame_column_types(data_frame):
    data_frame_col_dict = {}
    for col in data_frame:
        data_frame_col_dict[col] = type(data_frame[col][0]).__name__
    return data_frame_col_dict


def get_cql_columns_from_dataframe(data_frame):
    """
    Extracts a dictionary of column names and their cassandra data types from the dataframe
    Args:
        data_frame (DataFrame): the dataframe whose columns need to be extracted
    Returns: dict
    """
    column_dtype = _get_data_frame_column_types(data_frame)
    return {key: _PYTHON_TO_CASSANDRA_DATA_TYPE_MAP[value] for (key, value) in column_dtype.items()}


def _pandas_factory(colnames, rows):
    return pd.DataFrame(rows, columns=colnames)


def _chunk_list(lst: list, size: int) -> list:
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), size):
        yield lst[i:i + size]


def dataframe_to_cassandra_tuples(dataframe: DataFrame) -> list:
    """
    The Cassandra api uses tuples to send data for it's prepared statements, this method converts
    the rows of a dataframe into a list of tuples. It also converts the Pandas specific datatypes
    like Timestamp and NaN to python datatypes like datetime and None
    Args:
        dataframe (DataFrame): the dataframe to be converted to a list of tuples
    Returns: list[tuple]
    """
    return [tuple([_standardize_datatype(val) for val in row[1:]]) for row in dataframe.itertuples()]


def dicts_to_cassandra_tuples(data: list) -> list:
    """
    The Cassandra api uses tuples to send data for it's prepared statements, this method converts
    the list of dictionaries into a list of tuples
    Args:
        data (list[dict]): the list of dictonaries to be converted into a list of tuples
    Returns: list[tuple]
    """
    return [tuple(dct.values()) for dct in data]


def _standardize_datatype(val):
    if isinstance(val, Timestamp):
        return val.to_pydatetime()
    if val is NaT:
        return None
    return val


class ValidationError(Exception):
    """
    Exception class to raise validation issues
    Args:
        message (str): the error message
    """

    def __init__(self, message):
        super().__init__(message)


def _cql_manage_column_lists(data_frame, primary_key_column_list):
    column_dict = get_cql_columns_from_dataframe(data_frame)
    column_list = [f"{k} {v}" for (k, v) in column_dict.items()]
    _validate_primary_key_list(column_dict, primary_key_column_list)
    return column_list


def _validate_primary_key_list(column_dict, primary_key_column_list):
    if primary_key_column_list is None or not primary_key_column_list:
        raise ValidationError("please provide at least one primary key column")
    for key in primary_key_column_list:
        if key not in column_dict.keys():
            raise ValidationError(
                f"The column {key} is not in the column list, it cannot be specified as a primary "
                "key",
            )


class CassandraSecretsManager(SecretsManager):
    """
    Secrets manager for Cassandra configuration
    Args:
        source (KeyValueSource): a kv source that has secrets
        username_var (str): the variable name or the key for finding username
        password_var (str): the variable name or the key for finding password
    """

    def __init__(self, source: KeyValueSource = ENVIRONMENT,
                 username_var: str = "CASSANDRA_USERNAME",
                 password_var: str = "CASSANDRA_PASSWORD"):
        super().__init__([username_var, password_var], source)

        self.username = self.get_secret(username_var)
        self.password = self.get_secret(password_var)


@dataclass
class CassandraConnectionSettings:
    """Encapsulates the Cassandra connection settings"""
    cluster_ips: list
    port: int
    load_balancing_policy: LoadBalancingPolicy
    secrets_manager: CassandraSecretsManager
    ssl_options: dict = None


class CassandraConnectionManager:
    """
    Creates and manages connection to the cassandra cluster.
    Example -
    >>> from cassandra.policies import DCAwareRoundRobinPolicy
    >>> from cassandra.cqlengine import columns
    >>> from cassandra.cqlengine.management import sync_table
    >>> from cassandra.cqlengine.models import Model

    >>> load_balancing_policy = DCAwareRoundRobinPolicy(local_dc='AWS_VPC_AP_SOUTHEAST_2')

    >>> conn = CassandraConnectionManager(
    ...     settings = CassandraConnectionSettings(
    ...         cluster_ips=["1.1.1.1", "2.2.2.2"],
    ...         port=9042,
    ...         load_balancing_policy=load_balancing_policy,
    ...     )
    ... )

    >>> conn = CassandraConnectionManager(
    ...     CassandraConnectionSettings(
    ...         cluster_ips=["1.1.1.1", "2.2.2.2"],
    ...         port=9042,
    ...         load_balancing_policy=load_balancing_policy,
    ...         secrets_manager=CassandraSecretsManager(
    ...         username_var="MY_CUSTOM_USERNAME_ENV_VAR"),
    ...     )
    ... )

    For running Cassandra model operations
    >>> conn.setup_connection("dev_space")
    >>> class ExampleModel(Model):
    ...     example_type    = columns.Integer(primary_key=True)
    ...     created_at      = columns.DateTime()
    ...     description     = columns.Text(required=False)
    >>> sync_table(ExampleModel)

    Args:
        settings (hip_data_tools.apache.cassandra.CassandraConnectionSettings): settings to use
        for connecting to a cluster
    """

    def __init__(self, settings: CassandraConnectionSettings):
        self._settings = settings
        self.cluster = None
        self.session = None
        self._auth = PlainTextAuthProvider(
            username=self._settings.secrets_manager.username,
            password=self._settings.secrets_manager.password,
        )

    def get_cluster(self) -> Cluster:
        """
        get the cassandra Cluster object if it already exists or create a new one
        Returns: Cluster
        """
        if self.cluster is None:
            self.cluster = Cluster(
                contact_points=self._settings.cluster_ips,
                load_balancing_policy=self._settings.load_balancing_policy,
                port=self._settings.port,
                auth_provider=self._auth,
                ssl_options=self._settings.ssl_options,
            )
        return self.cluster

    def get_session(self, keyspace) -> Session:
        """
        get the cassandra Cluster's Session object if it already exists or create a new one
        Returns: Session
        """
        if self.session is None:
            self.session = self.get_cluster().connect(keyspace)

        return self.session

    def setup_connection(self, default_keyspace) -> None:
        """
        setups an implicit connection object for cassandra using the cassandra settings in the
        connection manager
        Args:
            default_keyspace (str): the keyspace to use as default for this implicit connection
        Returns: None
        """
        connection.setup(
            self._settings.cluster_ips,
            load_balancing_policy=self._settings.load_balancing_policy,
            auth_provider=self._auth,
            port=self._settings.port,
            ssl_options=self._settings.ssl_options,
            default_keyspace=default_keyspace)


class CassandraUtil:
    """
    Class to connect to and then retrieve, transform and upload data from and to cassandra
    """

    def __init__(self, keyspace: str, conn: CassandraConnectionManager):
        self.keyspace = keyspace
        self._conn = conn
        self.consistency_level = ConsistencyLevel.QUORUM
        self._session = self._conn.get_session(self.keyspace)

    def _cql_upsert_from_dict(self, data: dict, table: str):
        upsert_sql = f"""
        INSERT INTO {self.keyspace}.{table} 
        ({", ".join(data)}) 
        VALUES ({", ".join(['?' for key in data])});
            """
        LOG.debug(upsert_sql)
        return upsert_sql

    def _cql_upsert_from_dataframe(self, dataframe: DataFrame, table: str):
        upsert_sql = f"""
        INSERT INTO {self.keyspace}.{table} 
        ({", ".join(list(dataframe.columns.values))}) 
        VALUES ({", ".join(['?' for key in dataframe.columns.values])});
            """
        LOG.debug(upsert_sql)
        return upsert_sql

    def upsert_dataframe_in_batches(self,
                                    dataframe: DataFrame,
                                    table: str,
                                    batch_size: int = 2
                                    ) -> List[Result]:
        """
        Upload all data from a DataFrame onto a cassandra table
        Args:
            dataframe (DataFrame): a DataFrame to upsert
            table (str): the table to upsert data into
            table. If None then the DataFrame column names that match cassandra table anme will be
            upserted else ignored
            batch_size (int): limit on the number of prepared statements in the batch
        Returns: ResultSet
        """
        prepared_statement = self._session.prepare(
            self._cql_upsert_from_dataframe(dataframe, table))

        batches = self.prepare_batches(prepared_statement, dataframe_to_cassandra_tuples(dataframe),
                                       batch_size)
        return self._execute_batches(batches)

    def upsert_dataframe(self,
                         dataframe: DataFrame,
                         table: str
                         ) -> List[Result]:
        """
        Upload all data from a DataFrame onto a cassandra table
        Args:
            dataframe (DataFrame): a DataFrame to upsert
            table (str): the table to upsert data into
            table. If None then the DataFrame column names that match cassandra table anme will be
            upserted else ignored
            batch_size (int): limit on the number of prepared statements in the batch
        Returns: ResultSet
        """
        prepared_statement = self._session.prepare(
            self._cql_upsert_from_dataframe(dataframe, table))
        data_tuples = dataframe_to_cassandra_tuples(dataframe)
        return [self._execute_prepared_statement(data_tuple, prepared_statement)
                for data_tuple in data_tuples]

    def _execute_prepared_statement(self, data_tuple, prepared_statement):
        return self.execute(prepared_statement.bind(data_tuple), row_factory=dict_factory)

    def _execute_batches(self, batches: List):
        results = []
        LOG.info("Executing cassandra batches")
        for batch in batches:
            results.append(self._execute_batch(batch))
        LOG.info("finished %s batches", len(results))
        return results

    @retry(wait_exponential_multiplier=_RETRY_WAIT_MULTIPLIER_MS,
           wait_exponential_max=_RETRY_WAIT_MAX_MS,
           )
    def _execute_batch(self, batch):
        LOG.debug("Executing query: %s", batch)
        return self._session.execute(batch, timeout=300.0)

    def upsert_dictonary_list_in_batches(self,
                                         data: List[dict],
                                         table: str,
                                         batch_size: int = 2
                                         ) -> List[Result]:
        """
        Upsert a row into the cassandra table based on the dictionary key values
        Args:
            data (list[dict]): the data to be upserted
            table (str): the table to upsert data into
            batch_size (int): limit on the number of prepared statements in the batch
        Returns: None
        """
        prepared_statement = self._session.prepare(self._cql_upsert_from_dict(data[0], table))
        batches = self.prepare_batches(prepared_statement, dicts_to_cassandra_tuples(data),
                                       batch_size)
        return self._execute_batches(batches)

    def upsert_dictonary_list(self,
                              data: List[dict],
                              table: str,
                              ) -> List[Result]:
        """
        Upsert a row into the cassandra table based on the dictionary key values
        Args:
            data (list[dict]): the data to be upserted
            table (str): the table to upsert data into
        Returns: None
        """
        prepared_statement = self._session.prepare(self._cql_upsert_from_dict(data[0], table))
        data_tuples = dicts_to_cassandra_tuples(data)
        return [self._execute_prepared_statement(data_tuple, prepared_statement)
                for data_tuple in data_tuples]

    def create_table_from_model(self, model_class):
        """
        Create a table if not exists from the Model Class
        Args:
            model_class (class): class for the Model
        Returns: None
        """
        self._conn.setup_connection(default_keyspace=self.keyspace)
        sync_table(model_class)

    def create_table_from_dataframe(self,
                                    data_frame: DataFrame,
                                    table_name: str,
                                    primary_key_column_list: List[str],
                                    table_options_statement=""):
        """
        Create a new table in cassandra based on a pandas DataFrame
        Args:
            data_frame (DataFrame): the data frame to be synced
            table_name (str): name of the table to create
            primary_key_column_list (lost[str]): list of columns in the data frame that constitute
            primary key for new table
            table_options_statement (str): a cql valid WITH statement to specify table options as
            specified in https://docs.datastax.com/en/dse/6.0/cql/cql/cql_reference/cql_commands
            /cqlCreateTable.html#table_optionsŒ

        Returns: ResultSet

        """
        cql = self._dataframe_to_cassandra_ddl(
            data_frame,
            primary_key_column_list,
            table_name,
            table_options_statement)
        return self.execute(cql, row_factory=dict_factory)

    def _dataframe_to_cassandra_ddl(self,
                                    data_frame: DataFrame,
                                    primary_key_column_list: List[str],
                                    table_name: str,
                                    table_options_statement: str = ""):
        column_list = _cql_manage_column_lists(data_frame, primary_key_column_list)
        cql = f"""
        CREATE TABLE IF NOT EXISTS {self.keyspace}.{table_name} (
            {", ".join(column_list)},
            PRIMARY KEY ({", ".join(primary_key_column_list)}))
        {table_options_statement};
        """
        LOG.debug(cql)
        return cql

    def read_as_dictonary_list(self, query: str, **kwargs) -> List[dict]:
        """
        Read the results of a query in form of a list of dict
        Args:
            query (str):
        Returns: list[dict]
        """
        return self.execute(query, dict_factory, **kwargs).current_rows

    def read_as_dataframe(self, query: str, **kwargs) -> DataFrame:
        """
        Read the result of a query in form of a pandas DataFrame
        Args:
            query (str):
        Returns: DataFrame
        """
        # Since pandas DataFrame does not have a boolean truth value will need to access
        # protected _current_rows
        return self.execute(query, _pandas_factory, **kwargs)._current_rows

    def execute(self, query: str, row_factory: callable, **kwargs) -> Result:
        """
        Execute a cql command and retrieve data with the row factory
        Args:
            query (str):
            row_factory (callable):
            **kwargs: Kwargs to match the session.execute command in cassandra
        Returns: ResultSet
        """
        LOG.debug("Executing query: %s", query)
        if row_factory is not None:
            self._session.row_factory = row_factory
        return self._session.execute(query, **kwargs)

    def prepare_batches(self,
                        prepared_statement: PreparedStatement,
                        tuples: List[tuple],
                        batch_size: int) -> List:
        """
        Prepares a list of cassandra batched Statements out of a list of tuples and prepared
        statement
        Args:
            prepared_statement (PreparedStatement): the statement to be used for batching.
            tuples (list[tuple]): the data to be inserted.
            batch_size (int): limit on the number of prepared statements in the batch.
        Returns: list[BatchStatement]
        """
        batches = []
        LOG.debug("Preparing cassandra batches out of rows")
        batches_of_tuples = _chunk_list(tuples, batch_size)
        for tpl in batches_of_tuples:
            batch = self._prepare_batch(prepared_statement, tpl)
            batches.append(batch)
        LOG.info("created %s batches out of list of %s tuples", len(batches), len(tuples))
        return batches

    def _prepare_batch(self, prepared_statement, tuples):
        batch = BatchStatement(consistency_level=self.consistency_level)
        for tpl in tuples:
            batch.add(prepared_statement, tpl)
        return batch



"""
handle ETL of data from Athena to Athena
"""
from typing import Optional, List

from attr import dataclass

from hip_data_tools.aws.athena import AthenaUtil
from hip_data_tools.aws.common import AwsConnectionSettings, AwsConnectionManager


@dataclass
class AthenaToAthenaSettings:
    """Athena To Athena ETL settings"""
    source_sql: str
    source_database: str
    target_database: str
    target_table: str
    target_data_format: str
    target_s3_bucket: str
    target_s3_dir: str
    target_partition_columns: Optional[List[str]]
    connection_settings: AwsConnectionSettings


class AthenaToAthena:
    """
    ETL To transfer data from an Athena SQL into an Athena Table
    Args:
        settings (AthenaToAthenaSettings): Settings for the ETL
    """

    def __init__(self, settings: AthenaToAthenaSettings):
        self.__settings = settings
        self._athena = None

    def generate_create_table_statement(self) -> str:
        """
        Generates an Athena compliant ctas sql statement for the ETL
        Returns: str
        """
        partition_statement = ""
        if self.__settings.target_partition_columns:
            col_list = ','.join([f"'{c}'" for c in self.__settings.target_partition_columns])
            partition_statement = f", partitioned_by = ARRAY[{col_list}] "
        external_location = f"'s3://{self.__settings.target_s3_bucket}/" \
                            f"{self.__settings.target_s3_dir}/'"
        return f"""
            CREATE TABLE {self.__settings.target_database}.{self.__settings.target_table}
            WITH (
                format = '{self.__settings.target_data_format}'
                , external_location = {external_location}
                {partition_statement}
            ) AS 
            {self.__settings.source_sql}
            """

    def _get_athena_util(self) -> AthenaUtil:
        if self._athena is None:
            import uuid
            self._athena = AthenaUtil(
                database=self.__settings.source_database,
                conn=AwsConnectionManager(self.__settings.connection_settings),
                output_key=f"athena_results/{uuid.uuid4().hex}",
                output_bucket=self.__settings.target_s3_bucket,
            )
        return self._athena

    def execute(self) -> None:
        """
        Execute the ETL by running an Athena CTAS statement
        Returns: None
        """
        self._get_athena_util().run_query(self.generate_create_table_statement())



"""
handle ETL of data from Athena to S3
"""
from typing import Optional, List

from attr import dataclass

from hip_data_tools.aws.athena import AthenaUtil
from hip_data_tools.aws.common import AwsConnectionSettings, AwsConnectionManager
from hip_data_tools.etl.athena_to_athena import AthenaToAthenaSettings, AthenaToAthena
from hip_data_tools.etl.common import current_epoch, get_random_string


@dataclass
class AthenaToS3Settings:
    """Athena To S3 ETL settings"""
    source_sql: str
    source_database: str
    temporary_database: Optional[str]
    temporary_table: Optional[str]
    target_data_format: str
    target_s3_bucket: str
    target_s3_dir: str
    target_partition_columns: Optional[List[str]]
    connection_settings: AwsConnectionSettings


class AthenaToS3:
    """
    ETL To transfer data from an Athena sql to an s3 location
    Args:
        settings (AthenaToS3Settings): Settings for the etl
    """

    def __init__(self, settings: AthenaToS3Settings):
        self.__settings = settings
        if self.__settings.temporary_database:
            self.__settings.temporary_database = self.__settings.source_database
        if self.__settings.temporary_table:
            self.__settings.temporary_table = f"temp__{current_epoch()}__{get_random_string(5)}"

    def _drop_temporary_table(self) -> None:
        au = AthenaUtil(
            database=self.__settings.temporary_database,
            conn=AwsConnectionManager(self.__settings.connection_settings)
        )
        au.drop_table(self.__settings.temporary_table)

    def execute(self) -> None:
        """
        Execute the ETL to transfer data from Athena to S3
        Returns: None
        """
        etl = AthenaToAthena(
            AthenaToAthenaSettings(
                source_sql=self.__settings.source_sql,
                source_database=self.__settings.source_database,
                target_database=self.__settings.temporary_database,
                target_table=self.__settings.temporary_table,
                target_data_format=self.__settings.target_data_format,
                target_s3_bucket=self.__settings.target_s3_bucket,
                target_s3_dir=self.__settings.target_s3_dir,
                target_partition_columns=self.__settings.target_partition_columns,
                connection_settings=self.__settings.connection_settings
            )
        )
        etl.execute()
        # Delete temporary table
        self._drop_temporary_table()



"""
Module to deal with data transfer from S3 to Cassandra
"""
from typing import Iterator, List

import pandas as pd
from attr import dataclass
from pandas import DataFrame

from hip_data_tools.aws.common import AwsConnectionSettings, AwsConnectionManager
from hip_data_tools.aws.s3 import S3Util
from hip_data_tools.common import LOG


@dataclass
class S3ToDataFrameSettings:
    """S3 to Cassandra ETL settings"""
    source_bucket: str
    source_key_prefix: str
    source_connection_settings: AwsConnectionSettings


class S3ToDataFrame:
    """
    Class to transfer parquet data from s3 to Pandas DataFrame
    Args:
        settings (S3ToDataFrameSettings): the settings around the etl to be executed
    """

    def __init__(self, settings: S3ToDataFrameSettings):
        self._iterator = None
        self.__settings = settings
        self.keys_to_transfer = None
        self._processed_counter = 0

    def __iter__(self):
        self._processed_counter = 0
        return self

    def __next__(self) -> DataFrame:
        source_files = self.list_source_files()
        if source_files:
            data = self.get_data_frame(source_files[self._processed_counter])
            self._processed_counter += 1
            return data
        else:
            return DataFrame()

    def _get_s3_util(self) -> S3Util:
        return S3Util(
            bucket=self.__settings.source_bucket,
            conn=AwsConnectionManager(self.__settings.source_connection_settings),
        )

    def _get_iterator(self) -> Iterator:
        if self._iterator is None:
            self._iterator = iter(self)
        return self._iterator

    def get_all_files_as_data_frame(self) -> DataFrame:
        """
        Downloads and collates all files in a given s3 dir and returns a single DataFrame
        Returns: DataFrame
        """
        return pd.concat([self.get_data_frame(key) for key in self.list_source_files()])

    def get_data_frame(self, key: str) -> DataFrame:
        """
        Read a parquet file from s3 and convert it to a parquet DataFrame
        Args:
            key: s3 key for the parquet file
        Returns: None
        """
        return self._get_s3_util().download_parquet_as_dataframe(key=key)

    def list_source_files(self) -> List[str]:
        """
        Lists all the files that are encompassed under the s3 location in settings
        Returns: list[str]
        """
        if self.keys_to_transfer is None:
            self.keys_to_transfer = self._get_s3_util().get_keys(
                self.__settings.source_key_prefix)
            LOG.info("Listed and cached %s source files", len(self.keys_to_transfer))
        return self.keys_to_transfer

    def next(self) -> DataFrame:
        """
        Gets the next DataFrame from the next file on s3.

        Please note, if you are trying to run window functions or operations on the data set that
        spans multiple rows, then using this method may result in incorrect or inaccurate
        results. For such use cases, use `get_all_files_as_data_frame()`
        Returns: DataFrame
        """
        return next(self._get_iterator())



"""
Utility to connect to, and interact with the s3 file storage system
"""
import json
import os
import uuid
from multiprocessing.pool import Pool
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import List, Any

import arrow
import pandas as pd
from joblib import load, dump
from pandas import DataFrame

from hip_data_tools.aws.common import AwsUtil, AwsConnectionManager, AwsConnectionSettings
from hip_data_tools.common import _generate_random_file_name, LOG

UTF8 = 'utf-8'


class S3Util(AwsUtil):
    """
    Utility class for connecting to s3 and manipulate data in a python way
    Args:
        conn (AwsConnection): AwsConnection object or a boto.Session object
        bucket (string): S3 bucket name where these operations will take place
    """

    def __init__(self, conn: AwsConnectionManager, bucket: str):
        super().__init__(conn, "s3")
        self.bucket = bucket

    def download_file(self, local_file_path: str, key: str) -> None:
        """
        Downloads a file from a path on s3 to a local path on disk
        Args:
            local_file_path (str): Absolute path on s3.
            key (str): Absolute path to save file to local.
        Returns: None
        """
        self.get_client().download_file(self.bucket, key, local_file_path)

    def upload_file(self, local_file_path: str, key: str, remove_local: bool = True) -> None:
        """
        Uploads a file from local to s3
        Args:
            local_file_path (str): Absolute local path to the file to upload
            key (str): Absolute path within the s3 buck to upload the file
            remove_local (bool): remove file from local fs after transfer
        Returns: None
        """
        self.get_client().upload_file(local_file_path, self.bucket, key)
        if remove_local:
            os.remove(local_file_path)

    def download_object_and_deserialise(self, key: str, local_file_path: str = None):
        """
        Download a serialised object from S3 and deserialize
        Args:
            key (str): Absolute path on s3 to the file
            local_file_path (str): The deserialized object
        Returns: object
        """
        if local_file_path is None:
            local_file_path = "/tmp/tmp_file{}".format(str(uuid.uuid4()))

        self.download_file(key=key, local_file_path=local_file_path)
        return load(local_file_path)

    def serialise_and_upload_object(self, obj: Any, key: str) -> None:
        """
        Serialise any object to disk, and then upload to S3
        Args:
            obj (Any): Any serialisable object
            key (str): The absolute path on s3 to upload the file to
        Returns: None
        """

        random_tmp_file_nm = _generate_random_file_name()
        dump(obj, random_tmp_file_nm)
        self.upload_file(local_file_path=random_tmp_file_nm, key=key)

    def create_bucket(self) -> None:
        """
        Creates the s3 bucket
        Returns: None
        """
        self.get_resource().create_bucket(Bucket=self.bucket)

    def upload_dataframe_as_parquet(self,
                                    dataframe: DataFrame,
                                    key: str,
                                    file_name: str = "data") -> None:
        """
        Exports a datafame to a parquet file on s3
        Args:
            dataframe (DataFrame): dataframe to export
            key (str): The path on s3 to upload the file to (excluding bucket name and file name)
            file_name (str): the name of the file at destination
        Returns: None
        """
        LOG.debug(
            "Uploading the dataframe as parquet\nColumn names of the dataframe: %s\nTop 2 rows of "
            "the dataframe: %s\nShape of the dataframe: %s",
            list(dataframe), dataframe.head(2), dataframe.shape)

        tmp_file = NamedTemporaryFile(delete=False)
        destination = f"{key}/{file_name}.parquet"
        dataframe.to_parquet(fname=tmp_file.name)
        self.upload_file(local_file_path=tmp_file.name, key=destination, remove_local=True)

    def download_parquet_as_dataframe(self,
                                      key: str,
                                      engine: str = 'auto',
                                      columns: List[str] = None,
                                      **kwargs) -> DataFrame:
        """
        Exports a datafame to a parquet file on s3
        Args:
            key (str): The absolute path on s3 to upload the file to
            engine (str): parquet engine
            columns (lis[str]): list of columns default None to extrapolate from dataframe
        Returns: DataFrame
        """
        random_tmp_file_nm = _generate_random_file_name()
        self.download_file(random_tmp_file_nm, key)
        return pd.read_parquet(random_tmp_file_nm, engine=engine, columns=columns, **kwargs)

    def read_lines_as_list(self, key_prefix: str, encoding: str = UTF8) -> List[str]:
        """
        Read lines from s3 files
        Args:
            key_prefix (str): the key prefix under which all files will be read
            encoding: the character encoding to use for encoding / decoding content
        Returns: list[str] lines read from all files
        """
        s3 = self.get_resource()
        bucket = s3.Bucket(name=self.bucket)
        lines = []
        LOG.info("reading files from s3://%s/%s ", self.bucket, key_prefix)
        file_metadata = bucket.objects.filter(Prefix=key_prefix)
        for file in file_metadata:
            obj = s3.Object(self.bucket, file.key)
            data = obj.get()["Body"].read().decode(encoding)
            lines.append(data.splitlines())
        # Flatten the list of lists
        flat_lines = [item for sublist in lines for item in sublist]
        LOG.info("Read %d lines from %d s3 files", len(flat_lines), len(lines))
        return flat_lines

    def delete_recursive(self, key_prefix: str) -> None:
        """
        Recursively delete all keys with given prefix from the named bucket
        Args:
            key_prefix (str): Key prefix under which all files will be deleted
        Returns: NA
        """
        if not key_prefix.endswith("/"):
            key_prefix = f"{key_prefix}/"
        LOG.info("Recursively deleting s3://%s/%s", self.bucket, key_prefix)
        response = self.get_resource().Bucket(self.bucket).objects.filter(
            Prefix=key_prefix).delete()
        LOG.info(response)

    def get_keys(self, key_prefix: str) -> List[str]:
        """
        returns a list of all objects unser a given key prefix
        Args:
            key_prefix (str): Key Prefix under which all objects are to be listed
        Returns: list[str]
        """
        continuation_token = None
        keys = []
        while True:
            result = self._list_object_page(key_prefix, continuation_token)

            keys = keys + [content.get('Key', None) for content in result.get('Contents', [])]
            if 'NextContinuationToken' not in result:
                break
            continuation_token = result['NextContinuationToken']
        return keys

    def _list_object_page(self, key_prefix: str, continuation_token: str):
        if continuation_token is None:
            return self.get_client().list_objects_v2(
                Bucket=self.bucket,
                Prefix=key_prefix,
            )
        return self.get_client().list_objects_v2(
            Bucket=self.bucket,
            Prefix=key_prefix,
            ContinuationToken=continuation_token,
        )

    def upload_directory(self,
                         source_directory: str,
                         extension: str,
                         target_key: str,
                         overwrite: bool = True,
                         rename: bool = True) -> None:
        """
        Upload a local file directory to s3
        Args:
            source_directory (str): Local source directory's absolute path.
            extension (str): the file extension of files in that directory to be uploaded.
            target_key (str): Target location on the s3 bucket for files to be uploaded.
            overwrite (bool): overwrites files on s3 if set to true.
            rename (bool): renames the file when uploading to s3 if set to true.
        Returns: None
        """
        if overwrite:
            LOG.info("Cleaning existing files on s3")
            self.delete_recursive(f"{target_key}/")
        LOG.info("searching for files to upload in %s", source_directory)
        path_list = Path(source_directory).glob(f'**/*.{extension}')
        itr = 0
        upload_data = []
        for path in path_list:
            path_in_str = str(path)
            filename = os.path.basename(path_in_str)
            if rename:
                filename = f"file-{str(uuid.uuid4())}.{extension}"
            destination_key = f"{target_key}/{filename}"
            itr = itr + 1
            upload_data += [(self.conn.settings, path_in_str, self.bucket, destination_key)]
        Pool().starmap(_multi_process_upload_file, upload_data)  # Use all available cores
        LOG.info("Upload of directory complete at s3://%s/%s", self.bucket, target_key)

    def delete_recursive_match_suffix(self, key_prefix: str, suffix: str) -> None:
        """
        Recursively delete all keys with given key prefix and suffix from the bucket
        Args:
            key_prefix (str): Key prefix under which all files will be deleted.
            suffix (str): suffix of the subset of files in the given prefix directory to be deleted
        Returns: None
        """
        if not key_prefix:
            raise ValueError("key_prefix must not be empty")
        if not suffix:
            raise ValueError("suffix must not be empty")
        s3 = self.get_resource()
        for obj in s3.Bucket(self.bucket).objects.filter(Prefix=key_prefix):
            if obj.key.endswith(suffix):
                LOG.info("deleting s3://%s/%s", self.bucket, obj.key)
                response = obj.delete()
                LOG.info("Response: %s ", response)

    def download_directory(self, source_key: str, file_suffix: str, local_directory: str) -> None:
        """
        Download an entire directory from s3 onto local file system
        Args:
            source_key (str): key prefix of the directory to be downloaded from s3
            file_suffix (str): suffix to filter a subset under the source_key to be downloaded
            local_directory (str): local absolute path to store all the files
        Returns: None
        """
        s3 = self.get_resource()
        LOG.info("Downloading s3://%s/%s to %s", self.bucket, source_key, local_directory)
        for obj in s3.Bucket(self.bucket).objects.filter(Prefix=source_key):
            key_path = obj.key.split("/")
            if obj.key.endswith(file_suffix):
                filename = f"{local_directory}/{key_path[-1]}"
                self.download_file(
                    local_file_path=filename,
                    key=obj.key)

    def upload_json(self, key: str, json_list: List[dict], encoding: str = UTF8) -> None:
        """
        Save the json/dict data structure onto s3 as a file without using temporary local files
        Args:
            key: target key of the file on s3
            json_list: a list of dictionaries that are saved as newline json in a file
            encoding: the character encoding to use for encoding / decoding content
        Returns: None
        """
        s3 = self.get_resource()
        s3.Object(self.bucket, key).put(
            Body=(bytes(json.dumps(json_list, indent=2).encode(encoding)))
        )

    def download_json(self, key: str, encoding: str = UTF8) -> dict:
        """
        Read a file in json format from s3
        Args:
            key: location of the file to read
            encoding: the character encoding to use for encoding / decoding content
        Returns: dict
        """
        s3 = self.get_resource()
        json_content = json.loads(
            s3.Object(self.bucket, key).get()['Body'].read().decode(encoding)
        )
        return json_content

    def download_strings(self, key: str, encoding: str = UTF8) -> List[str]:
        """
        Read lines from s3 files
        Args:
            key: the key for the file which contains strings
            encoding: the character encoding to use for encoding / decoding content
        Returns: List[str]
        """
        s3 = self.get_resource()
        obj = s3.Object(self.bucket, key)
        data = obj.get()['Body'].read().decode(encoding)
        lines = data.splitlines()
        return lines

    def get_keys_modified_in_range(self,
                                   key_prefix: str,
                                   start_date: arrow,
                                   end_date: arrow) -> List[str]:
        """
        Sense if there were any files changed or added in the given time period under the given key
        prefix and return a list of keys
        Args:
            key_prefix: the key prefix under which all files will be sensed
            start_date: start of the duration in which the s3 objects were modified
            end_date: end of the duration in which the s3 objects were modified
        Returns: List[str]
        """
        LOG.info("sensing files from s3://%s/%s \n between %s to %s", self.bucket, key_prefix,
                 start_date, end_date)
        metadata = self.get_object_metadata(key_prefix)
        lines = []
        for file in metadata:
            if start_date < arrow.get(file.last_modified) <= end_date:
                lines += [file.key]
        LOG.info("found %s s3 files changed", len(lines))
        return lines

    def get_all_keys(self, key_prefix: str) -> List[str]:
        """
        Sense all keys under a given key prefix
        Args:
            key_prefix (str): the key prefix under which all files will be sensed
        Returns: List[str]
        """
        LOG.info("sensing files from s3://%s/%s ", self.bucket, key_prefix)
        metadata = self.get_object_metadata(key_prefix)
        lines = [file.key for file in metadata]
        LOG.info("found %s s3 keys", len(lines))
        return lines

    def get_object_metadata(self, key_prefix: str) -> List:
        """
        Get metadata for all objects under a key prefix
        Args:
            key_prefix: the key prefix under which all files will be sensed
        Returns: List[metadata]
        """
        s3 = self.get_resource()
        bucket = s3.Bucket(name=self.bucket)
        metadata = bucket.objects.filter(Prefix=key_prefix)
        return metadata

    def upload_binary_stream(self, stream: bytes, key: str) -> None:
        """
        Upload a binary stream of data as an s3 object's body
        Args:
            stream: the stream of bytes to be uploaded
            key: s3 key at which this stream is to be uploaded
        Returns: None
        """
        s3 = self.get_resource()
        obj = s3.Object(self.bucket, key)
        obj.put(Body=stream)

    def move_recursive(self,
                       source_dir: str,
                       destination_dir: str,
                       delete_after_copy: bool = True) -> None:
        """
        recursively move files on s3 to a new location on the same bucket
        Args:
            source_dir: Source key prefix representing the directory to move
            destination_dir: destination key prefix
            delete_after_copy: removes the files from source after successful copy if set to true
        Returns: None
        """
        s3 = self.get_resource()
        bucket = s3.Bucket(self.bucket)
        for obj in bucket.objects.filter(Prefix=source_dir):
            # replace the prefix
            new_key = destination_dir + obj.key[len(source_dir):]
            LOG.info("Moving s3 object from : \n%s \nto: \n%s", obj.key, new_key)
            new_obj = bucket.Object(new_key)
            new_obj.copy({'Bucket': self.bucket, 'Key': obj.key})
        if delete_after_copy:
            self.delete_recursive(source_dir)

    def rename_file(self, key: str, new_file_name: str) -> None:
        """
        Rename a file on s3
        Args:
            key: Current key of the file
            new_file_name: target file name
        Returns: None
        """
        s3 = self.get_resource()
        full_new_file_path = key.rpartition('/')[0] + '/' + new_file_name
        LOG.info("Renaming source: %s to %s", key, full_new_file_path)
        s3.Object(self.bucket, full_new_file_path).copy_from(
            CopySource={'Bucket': self.bucket, 'Key': key})
        s3.Object(self.bucket, key).delete()

    def download_strings_from_directory(self, key_prefix: str, encoding: str = UTF8) -> List[str]:
        """
        Read lines from s3 files
        Args:
            key_prefix: the key prefix under which all files will be read
            encoding: the character encoding to use for encoding / decoding content
        Returns: List[str]
        """
        s3 = self.get_resource()
        bucket = s3.Bucket(name=self.bucket)
        lines = []
        LOG.info("reading files from s3://%s/%s", self.bucket, key_prefix)
        file_metadata = bucket.objects.filter(Prefix=key_prefix)
        for file in file_metadata:
            obj = s3.Object(self.bucket, file.key)
            data = obj.get()['Body'].read().decode(encoding)
            lines.append(data.splitlines())
        # Flatten the list of lists
        flat_lines = [item for sublist in lines for item in sublist]
        LOG.info("read %s lines from %s s3 files", len(flat_lines), len(lines))
        return flat_lines


def _multi_process_upload_file(settings: AwsConnectionSettings, filename: str, bucket: str,
                               key: str) -> None:
    """
    A standalone copy of the method making it simple to pickle in a multi processing pool
    Args:
        settings: the s3 connection settings to use for upload
        filename: local file name of the file to be uploaded.
        bucket: the s3 bucket to upload file to.
        key: the s3 key to use while uploading the file
    Returns: None
    """
    LOG.info("Uploading File %s to s3://%s/%s", filename, bucket, key)
    S3Util(
        conn=AwsConnectionManager(settings),
        bucket=bucket
    ).upload_file(local_file_path=filename, key=key)



"""
This Module handles the connection and operations on Google AdWords accounts using adwords API
"""
import gzip
from collections import OrderedDict
from tempfile import NamedTemporaryFile
from typing import List, Optional, Any

import pandas as pd
from attr import dataclass
from googleads import oauth2, AdWordsClient
from googleads.adwords import ServiceQueryBuilder, ReportQuery
from googleads.common import GoogleSoapService
from googleads.oauth2 import GoogleOAuth2Client
from pandas import DataFrame

from hip_data_tools.common import KeyValueSource, ENVIRONMENT, SecretsManager, \
    nested_list_of_dict_to_dataframe, LOG


class GoogleAdWordsSecretsManager(SecretsManager):
    """
    Secrets manager for Google Adwords Specific secrets, these are modeled to handle the OAuth
    based google client and refresh token secrets
    Args:
        source (KeyValueSource): The Key Value source which holds secrets
        client_secret_var (str): The Key pointer for client_secret
        refresh_token_var (str): The Key pointer for refresh_token
        developer_token_var (str): The Key pointer for developer_token
    """

    def __init__(self,
                 source: KeyValueSource = ENVIRONMENT,
                 client_secret_var: str = "adwords_client_secret",
                 refresh_token_var: str = "adwords_refresh_token",
                 developer_token_var: str = "adwords_developer_token"):
        self._required_keys = [client_secret_var, refresh_token_var, developer_token_var, ]
        super().__init__(self._required_keys, source)
        self.client_secret = self.get_secret(client_secret_var)
        self.refresh_token = self.get_secret(refresh_token_var)
        self.developer_key = self.get_secret(developer_token_var)


@dataclass
class GoogleOAuthConnectionSettings:
    """
    Handle connection settings for AdWords client
    """
    client_id: str
    secrets_manager: GoogleAdWordsSecretsManager


class GoogleOAuthConnectionManager:
    """
    Creates an OAuth client from google adwords library
    Args:
        settings (GoogleOAuthConnectionSettings): connection settings to use for creation
    """

    def __init__(self, settings: GoogleOAuthConnectionSettings):
        self.__settings = settings

    def _get_oauth_client(self) -> GoogleOAuth2Client:
        """
        Create an OAuth refresh-token client for all google adwords and manager api
        Returns: GoogleOAuth2Client
        """
        return oauth2.GoogleRefreshTokenClient(
            client_id=self.__settings.client_id,
            client_secret=self.__settings.secrets_manager.client_secret,
            refresh_token=self.__settings.secrets_manager.refresh_token
        )


@dataclass
class GoogleAdWordsConnectionSettings(GoogleOAuthConnectionSettings):
    """
    Handle connection settings for AdWords client
    """
    user_agent: str
    client_customer_id: Optional[str]


class GoogleAdWordsConnectionManager(GoogleOAuthConnectionManager):
    """
    AdWords Connection Manager that manages the lifecycle of authentication and the resulting
    adwords client
    Args:
        settings (GoogleAdWordsConnectionSettings): setting to use for connecting to adwords
    """

    def __init__(self, settings: GoogleAdWordsConnectionSettings):
        super().__init__(settings)
        self.settings = settings
        self._adwords_client = None

    def get_adwords_client(self, **kwargs) -> AdWordsClient:
        """
        Gets the connected  adwords client, and creates one if not exist
        Returns (AdWordsClient): Adwords Client

        """
        if self._adwords_client is None:
            self._adwords_client = self._create_client(**kwargs)
        return self._adwords_client

    def _create_client(self, **kwargs) -> AdWordsClient:
        oauth2_client = self._get_oauth_client()
        adwords_client = AdWordsClient(
            developer_token=self.settings.secrets_manager.developer_key,
            oauth2_client=oauth2_client,
            user_agent=self.settings.user_agent,
            client_customer_id=self.settings.client_customer_id,
            **kwargs)
        return adwords_client


class AdWordsUtil:
    """
    Generic Adwords Utility class generates the required services to consume adwords API
    Args:
        conn (GoogleAdWordsConnectionManager): Connection manager to handle the creation of
        adwords client
    """

    def __init__(self, conn: GoogleAdWordsConnectionManager, service: str, version: str):
        self.service = service
        self.version = version
        self.conn = conn
        self._service_object = None
        self._pager = None
        self.query = None

    def _get_service(self, **kwargs) -> GoogleSoapService:
        if not self._service_object:
            self._service_object = self._create_service(**kwargs)
        return self._service_object

    def _create_service(self, **kwargs) -> GoogleSoapService:
        client = self.conn.get_adwords_client(**kwargs)
        service = client.GetService(self.service, version=self.version)
        return service


class AdWordsDataReader(AdWordsUtil):
    """
    Generic Adwords Utility class generates the required services to consume adwords API
    Args:
        conn (GoogleAdWordsConnectionManager): Connection manager to handle the creation of
        adwords client
    """

    def __init__(self, conn: GoogleAdWordsConnectionManager, service: str, version: str,
                 page_size: int = 1000):
        super().__init__(conn, service, version)
        self.__pager = None
        self.query = None
        self.page_size = page_size

    def download_all_as_dict(self) -> List[dict]:
        """
        Generates list of Dict from the adwords query language query as described in adwords api
        Returns: List[dict]
        """
        complete_result = []
        for page in self._get_query_pager(self._get_query()):
            complete_result.extend(get_page_as_list_of_dict(page))
        return complete_result

    def download_all_as_dataframe(self) -> DataFrame:
        """
        Generates a data frame from the next page of the API call to adwords
        Returns: List[dict]
        """
        data_list = self.download_all_as_dict()
        return nested_list_of_dict_to_dataframe(data_list)

    def download_next_page_as_dict(self) -> List[dict]:
        """
        Generates a list of dict from the next page of the API call to adwords
        Returns: List[dict]
        """
        return get_page_as_list_of_dict(next(self._get_query_pager(self._get_query())))

    def download_next_page_as_dataframe(self) -> DataFrame:
        """
        Generates a data frame from the next page of the API call to adwords
        Returns: List[dict]
        """
        data_list = self.download_next_page_as_dict()
        return nested_list_of_dict_to_dataframe(data_list)

    def set_query(self, query: ServiceQueryBuilder) -> None:
        """
        Sets the new query for the class to use awql, once the query is set, the download methods
        can iterate over paged results
        Args:
            query (ServiceQueryBuilder): query built as per adwords query language
        Returns: None
        """
        self.__pager = None
        self.query = query

    def _get_query(self) -> ServiceQueryBuilder:
        if self.query is None:
            raise Exception("Please set the query attribute using the method set_query(...)")
        return self.query

    def _get_query_pager(self, query: ServiceQueryBuilder):
        if self.__pager is None:
            self.__pager = query.Pager(self._get_service())
        return self.__pager

    def get_parallel_payloads(self, page_size: int, number_of_workers: int) -> List[dict]:
        """
        gives a list of dicts that contain start index, page size, and number of iterations
        Args:
            page_size (int): number of elements in each page / api call
            number_of_workers (int): total number of parallel workers for which the payload needs
            to be distributed
        Returns: List[dict] eg:
        [
            {'number_of_pages': 393, 'page_size': 1000, 'start_index': 0, 'worker': 0},
            {'number_of_pages': 393, 'page_size': 1000, 'start_index': 393000, 'worker': 1},
            {'number_of_pages': 393, 'page_size': 1000, 'start_index': 786000, 'worker': 2},
        ]
        """
        estimator = AdWordsParallelDataReadEstimator(
            conn=self.conn, service=self.service, version=self.version, query=self._get_query())
        return estimator.get_parallel_payloads(page_size, number_of_workers)


class AdWordsParallelDataReadEstimator(AdWordsUtil):
    """
    Class to use a service endpoint and awql query to estimate the number of parallel payloads, and
    their offsets
    Args:
        conn (GoogleAdWordsConnectionManager): Connection manager to handle the creation of
        adwords client
        service (str): Adwords service api to use for querying
        version (str): Adwords service api version to use for querying
        query (ServiceQueryBuilder): the Query builder object without limit clause
    """

    def __init__(self, conn: GoogleAdWordsConnectionManager,
                 service: str,
                 version: str,
                 query: ServiceQueryBuilder):
        super().__init__(conn, service, version)
        self.query = query

    def get_parallel_payloads(self, page_size: int, number_of_workers: int) -> List[dict]:
        """
        gives a list of dicts that contain start index, page size, and number of iterations
        Args:
            page_size (int): number of elements in each page / api call
            number_of_workers (int): total number of parallel workers for which the payload needs
            to be distributed
        Returns: List[dict] eg:
        [
            {'number_of_pages': 393, 'page_size': 1000, 'start_index': 0, 'worker': 0},
            {'number_of_pages': 393, 'page_size': 1000, 'start_index': 393000, 'worker': 1},
            {'number_of_pages': 393, 'page_size': 1000, 'start_index': 786000, 'worker': 2},
        ]
        """
        total_entries = self._get_total_entries()
        last_page_residue = total_entries % page_size
        number_of_pages = int(total_entries / page_size)
        if last_page_residue > 0:
            number_of_pages += 1
        pages_per_worker = round(number_of_pages / number_of_workers)
        start_index = 0
        result = []
        for worker in range(number_of_workers):
            payload = {
                "worker": worker,
                "start_index": start_index,
                "number_of_pages": pages_per_worker,
                "page_size": page_size,
            }
            start_index = start_index + (page_size * pages_per_worker)
            result.append(payload)
        return result

    def _get_total_entries(self) -> int:
        query = self.query
        result = next(query.Pager(self._get_service()))
        if 'totalNumEntries' in result:
            return result['totalNumEntries']
        return 0


class AdWordsCustomerUtil(AdWordsDataReader):
    """
    Adwords Utility to handle customer details, customer here meaning, the Adwords customer
    account, ie. the accounts that you see on top right corner on the AdWords console
    Args:
        conn (GoogleAdWordsConnectionManager): Connection manager to handle the creation of
        adwords client
    """

    def __init__(self, conn: GoogleAdWordsConnectionManager):
        super().__init__(conn, service='CustomerService', version='v201809')

    def get_customers(self) -> List[dict]:
        """
        Gets the customer details of the adwords accounts associated with the connection
        Returns: List[dict]

        """
        return self._get_service().getCustomers()


def _find_and_append_data(fail: dict, data: List[dict]) -> dict:
    for element in fail['fieldPathElements']:
        if element['field'] == 'operations':
            fail_index = element['index']
            fail["data"] = data[fail_index]
            return fail
    return fail


class AdWordsOfflineConversionUtil(AdWordsUtil):
    """
    Adwords Utility to handle offline conversions
    Args:
        conn (GoogleAdWordsConnectionManager): Connection manager to handle the creation of
        adwords client
    """

    def __init__(self, conn: GoogleAdWordsConnectionManager):
        super().__init__(conn, service='OfflineConversionFeedService', version='v201809')
        self.required_fields = [
            'conversionName',
            'conversionTime',
            'conversionValue',
            'googleClickId',
        ]
        self.valid_fields = self.required_fields + [
            'conversionCurrencyCode',
            'externalAttributionCredit',
            'externalAttributionModel',
        ]
        self.conversion_time_format = '%Y%m%d %H%M%S %Z'

    def upload_conversions(self, data: List[dict]) -> (List[dict], List[dict]):
        """
        Upload a list of conversions as a batch
        Args:
            data List[dict]: a List of conversion dictionaries like:
            [
                {
                    'googleClickId': 'some valid gclid',
                    'conversionName': 'some conversion name',
                    'conversionTime': '20200309 074353 UTC',
                    'conversionValue': 17.0,
                    'conversionCurrencyCode': 'AUD',
                },
                {
                    'googleClickId': 'some other valid gclid',
                    'conversionName': 'some conversion name',
                    'conversionTime': '20200309 023001 UTC',
                    'conversionValue': 17.0,
                    'conversionCurrencyCode': 'AUD',
                    'externalAttributionCredit': 0.3,
                    'externalAttributionModel': 'Linear',
                },
            ]

        Returns (List[dict], List[dict]): A tuple of lists holding succeeded and failed records,
        respectively eg.
        (
            [
                {
                    'googleClickId': 'gclid that succeeded',
                    'conversionName': 'some conversions name',
                    'conversionTime': '20200309 074353 UTC',
                    'conversionValue': 17.0,
                    'conversionCurrencyCode': 'AUD',
                    'externalAttributionCredit': None,
                    'externalAttributionModel': None
                },
            ] ,
             [
                 {
                    'fieldPath': 'operations[0].operand',
                    'fieldPathElements': [
                        {
                            'field': 'operations',
                            'index': 0
                        },
                        {
                            'field': 'operand',
                            'index': None
                        }
                    ],
                    'trigger': None,
                    'errorString': 'OfflineConversionError.UNPARSEABLE_GCLID',
                    'ApiError.Type': 'OfflineConversionError',
                    'reason': 'UNPARSEABLE_GCLID',
                    'data': {
                        'googleClickId': 'Gclid that failed',
                        'conversionName': 'some conversion name',
                        'conversionTime': '20200309 074353 UTC',
                        'conversionValue': 17.0,
                        'conversionCurrencyCode': 'AUD',
                        'externalAttributionCredit': None,
                        'externalAttributionModel': None
                    }
                },
            ]
        )

        """
        if not data:
            return [], []
        else:
            mutations = self._get_mutations_from_conversions_batch(data)
            result = self._upload_mutations_batch(mutations)
            if result['ListReturnValue.Type'] != 'OfflineConversionFeedReturnValue':
                raise Exception(
                    f"Unhandled Exception while loading batch of conversions, response: {result}")
            uploaded = [x for x in result['value'] if x is not None]
            #  Append actual data to the failed conversions
            fails = [_find_and_append_data(fail, data) for fail in result['partialFailureErrors']]

            return uploaded, fails

    def _upload_mutations_batch(self, mutations: List[dict]) -> dict:
        return self._get_service(partial_failure=True).mutate(mutations)

    def _get_mutations_from_conversions_batch(self, conversions: List[dict]) -> List[dict]:
        return [self._get_mutation_from_conversion(d) for d in conversions]

    def _get_mutation_from_conversion(self, conversion: dict) -> dict:
        self._verify_required_columns(conversion)
        self._verify_accepted_columns(conversion)
        return {'operator': 'ADD', 'operand': conversion}

    def _verify_required_columns(self, conversion: dict) -> None:
        for col in self.required_fields:
            if col not in conversion:
                raise ValueError(
                    f"The column {col} is required byt not present in the data {conversion}")

    def _verify_accepted_columns(self, conversion: dict) -> None:
        for col in conversion.keys():
            if col not in self.valid_fields:
                raise ValueError(
                    f"The column {col} present in the DataFrame is not in the allowed column list "
                    f"{self.valid_fields}")


def get_page_as_list_of_dict(page: dict) -> List[OrderedDict]:
    """
    Converts a list of entries from google adwords response into a list of Ordered Dictionaries
    Args:
        page (dict): the response page from google adwords api
    Returns: List[dict]
    """
    result = []
    if 'entries' in page:
        entries = page['entries']
        # These entries are a list of zeep Objects that need conversion to Dict
        result = [zeep_object_to_dict(entry) for entry in entries]
        LOG.debug("The result from the adword API: %s", result)
    else:
        LOG.info('No entries were found.')
    return result


def zeep_object_to_dict(obj: Any) -> OrderedDict:
    """
    converts the zeep client objects used in google adwords to Ordered Dict, also converts nested
    objects to ordered dict recursively
    Args:
        obj (Any): the Zeep object to be converted
    Returns: OrderedDict
    """
    obj_dict = obj.__dict__["__values__"]
    for key, val in obj_dict.items():
        if "zeep.objects" in f"{type(val)}":
            obj_dict[key] = zeep_object_to_dict(val)
    return obj_dict


class AdWordsCampaignUtil(AdWordsDataReader):
    """
    Handles the querying of Adwords Campaign service
    Args:
        conn (GoogleAdWordsConnectionManager): Connection manager to handle the creation of
        adwords client
    """

    def __init__(self, conn: GoogleAdWordsConnectionManager):
        super().__init__(conn, service='CampaignService', version='v201809')

    def set_query_to_fetch_all(self, start_index: int = 0,
                               page_size: int = 500) -> None:
        """
        Get all campaigns associated with an account
        Returns: None
        """
        query = (ServiceQueryBuilder()
                 .Select('Id', 'Name', 'Status')
                 # .Where('Status').EqualTo('ENABLED')
                 .OrderBy('Id')
                 .Limit(start_index, page_size)
                 .Build())
        self.set_query(query)


class AdWordsAdGroupUtil(AdWordsDataReader):
    """
    Handles the querying of Adwords AdGroup service
    Args:
        conn (GoogleAdWordsConnectionManager): Connection manager to handle the creation of
        adwords client
    """

    def __init__(self, conn: GoogleAdWordsConnectionManager):
        super().__init__(conn=conn, service='AdGroupService', version='v201809')
        self._all_query = None

    def set_query_to_fetch_by_campaign(self, campaign_id: str, start_index: int = 0,
                                       page_size: int = 500) -> None:
        """
        Get all AdGroups for the given Campaign
        Args:
            campaign_id (str): the adwords campaign to query
            page_size:
            start_index:
        Returns: None
        """
        query = (ServiceQueryBuilder()
                 .Select('Id', 'CampaignId', 'CampaignName', 'Status')
                 .Where('CampaignId').EqualTo(campaign_id)
                 .OrderBy('Id')
                 .Limit(start_index, page_size)
                 .Build())
        self.set_query(query)

    def set_query_to_fetch_all(self, start_index: int = 0, page_size: int = 500) -> None:
        """
        Get all Ad groups in the account
        Args:\
            page_size:
            start_index:
        Returns: None
        """
        query = (ServiceQueryBuilder()
                 .Select('Id', 'CampaignId', 'CampaignName', 'Status', 'Settings', 'Labels',
                         'ContentBidCriterionTypeGroup', 'BaseCampaignId', 'BaseAdGroupId',
                         'TrackingUrlTemplate', 'FinalUrlSuffix', 'UrlCustomParameters',
                         'AdGroupType')
                 .OrderBy('Id').Limit(start_index, page_size)
                 .Build())
        self.set_query(query)


class AdWordsAdGroupAdUtil(AdWordsDataReader):
    """
    Handles the querying of Adwords AdGroupAd service
    Args:
        conn (GoogleAdWordsConnectionManager): Connection manager to handle the creation of
        adwords client
    """

    def __init__(self, conn: GoogleAdWordsConnectionManager):
        super().__init__(conn, service='AdGroupAdService', version='v201809')
        self._all_query = None

    def set_query_to_fetch_all(self, start_index: int = 0, page_size: int = 500):
        """
        Get all Ad groups Ads in the account
        Args:
            page_size:
            start_index:
        Returns: None
        """
        query = (ServiceQueryBuilder()
                 .Select('Id')
                 .OrderBy('Id')
                 .Limit(start_index, page_size)
                 .Build())
        self.set_query(query)


class AdWordsReportDefinitionReader(AdWordsUtil):
    """
    Class to interact with the ReportDefinitionService and access fields for adwords reports
    Args:
        conn (GoogleAdWordsConnectionManager): Connection manager to handle the creation of
    adwords client
    """

    def __init__(self, conn: GoogleAdWordsConnectionManager):
        super().__init__(conn=conn, service="ReportDefinitionService", version="v201809")

    def get_report_fields(self, report_type: str) -> List[dict]:
        """
        Get a list of fields for a given report.
        Args:
            report_type (str): Possible reports are defined in the documentation here -
            https://developers.google.com/adwords/api/docs/appendix/reports/all-reports
        Returns: List[dict]
        """
        return self._get_service().getReportFields(report_type)


class AdWordsReportReader:
    """
    Generic data reader class for downloading data from report awql efficiently
    Args:
            conn (GoogleAdWordsConnectionManager): Connection manager to handle the creation of
        adwords client
    """

    def __init__(self, conn: GoogleAdWordsConnectionManager):
        self.conn = conn
        self.version = 'v201809'
        self.__downloader = None

    def _get_report_downloader(self):
        if self.__downloader is None:
            client = self.conn.get_adwords_client()
            self.__downloader = client.GetReportDownloader(version=self.version)
        return self.__downloader

    def awql_to_dataframe(self, query: ReportQuery,
                          include_zero_impressions: bool = True,
                          **kwargs) -> DataFrame:
        """
        Download the data returned by report query in a compressed format and return it in form of
        a pandas dataframe
        Args:
            query (ReportQuery): an awql report query, see example
            https://github.com/googleads/googleads-python-lib/blob/master/examples/adwords
            /v201809/reporting/stream_criteria_report_results.py#L34-L39
            include_zero_impressions (bool): include_zero_impressions
        Returns: DataFrame
        """
        with NamedTemporaryFile(mode="w+b") as temp_file:
            self._get_report_downloader().DownloadReportWithAwql(
                query,
                "GZIPPED_CSV",
                temp_file,
                skip_report_header=True,
                skip_column_header=False,
                skip_report_summary=True,
                include_zero_impressions=include_zero_impressions,
                **kwargs
            )
            with gzip.open(temp_file.name, mode="rt") as csv_file:
                dataframe = pd.read_csv(csv_file, sep=",", header=0)
        return dataframe


class AdWordsManagedCustomerUtil(AdWordsUtil):
    """
    Adwords Utility to parse through and gather Account Information for all sub accounts
    Args:
        conn (GoogleAdWordsConnectionManager): Connection manager to handle the creation of
        adwords client
    """

    def __init__(self, conn: GoogleAdWordsConnectionManager):
        super().__init__(conn, service='ManagedCustomerService', version='v201809')
        self.selector_fields = ['Name', 'CustomerId', 'DateTimeZone', 'CurrencyCode',
                                'CanManageClients', 'TestAccount', 'AccountLabels']

    def _init_selector(self, offset, page_size):
        selector = {
            'fields': self.selector_fields,
            'paging': {
                'startIndex': str(offset),
                'numberResults': str(page_size)
            }
        }
        return selector

    def get_all_accounts(self, page_size: int = 1000) -> List[dict]:
        """
        Gets the customer details of the adwords accounts associated with the connection
        Returns: List[dict]
        """
        # Construct selector to get all accounts.
        offset = 0
        selector = self._init_selector(offset, page_size)
        more_pages = True
        all_accounts = []
        while more_pages:
            # Get serviced account graph.
            page = self._get_service().get(selector)
            all_accounts.extend(get_page_as_list_of_dict(page))
            offset += page_size
            selector['paging']['startIndex'] = str(offset)
            more_pages = offset < int(page['totalNumEntries'])

        return all_accounts

    def get_all_accounts_as_dataframe(self, page_size: int = 1000) -> DataFrame:
        """
        Gets the customer details of the adwords accounts associated with the connection
        Returns: DataFrame
        """
        return nested_list_of_dict_to_dataframe(self.get_all_accounts(page_size=page_size))



"""
Utility to connect to, and perform DML and DDL operations on aws Athena
"""

import csv
import sys
import time
from typing import List, Any, Tuple

from pandas import DataFrame

from hip_data_tools.aws.common import AwsUtil, AwsConnectionManager
from hip_data_tools.common import LOG

_PYTHON_TO_ATHENA_DATA_TYPE_MAP = {
    "Timestamp": "TIMESTAMP",
    "str": "STRING",
    "int64": "BIGINT",
    "int32": "INT",
    "dict": "MAP",
    "float64": "DOUBLE",
    "UUID": "STRING",
    "object": "STRING",
    "int": "INT",
    "list": "STRING",
    "NoneType": "STRING",
    "bool_": "BOOLEAN",
    "Bool": "BOOLEAN",
    "bool": "BOOLEAN",
}


def get_partitions_from_partitions_dict(partitions: dict):
    """
    Get the Athena table settings partitions list of dictionary
    Args:
        partitions (dict): dictionary of partition column name and value
    :return: List of dictionary
    """
    if partitions is not None:
        column_name = partitions["column"]
        return [{"column": column_name, "type": "STRING"}]
    return None


def extract_athena_type_from_value(val: Any) -> str:
    """
    Extracts athena compatible datatype from a given python object
    Args:
        val (Any): Any value from primitive python types
    Returns: str
    """
    return _PYTHON_TO_ATHENA_DATA_TYPE_MAP[type(val).__name__]


def get_table_settings_for_dataframe(dataframe: DataFrame, partitions: dict, s3_bucket: str,
                                     s3_dir: str, table: str) -> dict:
    """
    Get the Athena table settings
    Args:
        dataframe (DataFrame): data frame with column types and names
        partitions (dict): dictionary of partition column name and value
        s3_bucket (str): S3 bucket name to store Athena table's data
        s3_dir (str): S3 directory to store Athena table's data
        table (str): Name of the table
    :return: table settings
    """
    table_settings = {
        "exists": True,
        "partitions": get_partitions_from_partitions_dict(partitions),
        "storage_format_selector": "parquet",
        "encryption": False,
        "table": table,
        "columns": get_athena_columns_from_dataframe(dataframe),
        "s3_bucket": s3_bucket,
        "s3_dir": s3_dir,
    }
    return table_settings


class AthenaUtil(AwsUtil):
    """
    Utility class for connecting to athena and manipulate data in a pythonic way

    Args:
        database (string): the athena database to run queries on
        conn (AwsConnection): AwsConnection object
        output_key (string): the s3 key where the results of athena queries will be stored
        output_bucket (string): the s3 bucket where the results of athena queries will be stored
        work_group (string): Athena workGroup name
    """

    def __init__(self,
                 database: str,
                 conn: AwsConnectionManager,
                 output_key: str = None,
                 output_bucket: str = None,
                 work_group: str = "primary"):
        super().__init__(conn, "athena")
        self.database = database
        self.conn = conn
        self.output_key = output_key
        self.output_bucket = output_bucket
        self.work_group = work_group
        self.storage_format_lookup = {
            "parquet": {
                "row_format_serde": "org.apache.hadoop.hive.ql.io.parquet.serde.ParquetHiveSerDe",
                "outputformat": "org.apache.hadoop.hive.ql.io.parquet.MapredParquetOutputFormat",
                "inputformat": "org.apache.hadoop.hive.ql.io.parquet.MapredParquetInputFormat"
            },
            "csv": {
                "row_format_serde": "org.apache.hadoop.hive.serde2.OpenCSVSerde",
                "outputformat": "org.apache.hadoop.hive.ql.io.HiveIgnoreKeyTextOutputFormat",
                "inputformat": "org.apache.hadoop.mapred.TextInputFormat"
            },
            "json": {
                "row_format_serde": "org.openx.data.jsonserde.JsonSerDe",
                "outputformat": "org.apache.hadoop.hive.ql.io.HiveIgnoreKeyTextOutputFormat",
                "inputformat": "org.apache.hadoop.mapred.TextInputFormat"
            }
        }

    def run_query(self, query_string, return_result=False):
        """
        General purpose query executor that submits an athena query, then uses the execution id
        to poll and monitor the
        sucess of the query. and optionally return the result.
        Args:
            query_string (string): The string contianing valid athena query
            return_result (boolean): Boolean flag to turn on results

        Returns (boolean): if return_result = True then returns result dictionary, else None

        """
        athena = self.get_client()
        output_location = "s3://{bucket}/{key}".format(
            bucket=self.output_bucket,
            key=self.output_key)
        LOG.info("executing query \n%s \non database - %s with results location %s", query_string,
                 self.database,
                 output_location)
        response = athena.start_query_execution(
            QueryString=query_string,
            QueryExecutionContext={
                'Database': self.database
            },
            ResultConfiguration={
                'OutputLocation': output_location
            },
            WorkGroup=self.work_group
        )
        execution_id = response['QueryExecutionId']
        stats = self.watch_query(execution_id)
        LOG.info("athena response %s", response)
        if stats['QueryExecution']['Status']['State'] == 'SUCCEEDED':
            LOG.info("Query execution id - %s SUCCEEDED", execution_id)
            if return_result:
                return self._get_query_result(execution_id)
        else:
            raise ValueError("Query exited with {} state because {}".format(
                stats['QueryExecution']['Status']['State'],
                stats['QueryExecution']['Status']['StateChangeReason']))
        return None

    def watch_query(self, execution_id, poll_frequency=10):
        """
        Watch the query execution for a given execution id in Athena
        Args:
            execution_id: the execution id of an Athena Auery
            poll_frequency (int): Freq in seconds to poll for the query status using Athen API

        Returns: dictionary of status from Athena

        """
        LOG.info("Watching query with execution id - %s", execution_id)
        while True:
            athena = self.get_client()
            stats = athena.get_query_execution(QueryExecutionId=execution_id)
            status = stats['QueryExecution']['Status']['State']
            if status in ['SUCCEEDED', 'FAILED', 'CANCELLED']:
                LOG.info("Query Completed %s", stats)
                return stats
            time.sleep(poll_frequency)

    def _show_result(self, execution_id, max_result_size=1000):
        results = self._get_query_result(execution_id, max_result_size)
        column_info = results['ResultSet']['ResultSetMetadata']['ColumnInfo']
        headers = [h['Name'].encode('utf-8') for h in column_info]
        LOG.info(headers)
        csv_writer = csv.writer(sys.stdout, quoting=csv.QUOTE_ALL)
        csv_writer.writerows(
            [[val['VarCharValue'] for val in row['Data']] for row in results['ResultSet']['Rows']])

    def _get_query_result(self, execution_id, max_result_size=1000):
        athena = self.get_client()
        results = athena.get_query_results(QueryExecutionId=execution_id,
                                           MaxResults=max_result_size)
        # TODO: Add ability to parse pages larger than 1000 rows
        return results

    def repair_table_partitions(self, table):
        """
        Runs repair on given table
        Args:
            table (string): name of the table whose partitions need to be scanned and refilled

        Returns: None

        """
        self.run_query("MSCK REPAIR TABLE {}".format(table))

    def add_partitions(self, table, partition_keys, partition_values):
        """
        Add a new partition to a given table
        Args:
            table (string): name of the table to which a new partition is added
            partition_keys (list): an array of the keys/partition columns
            partition_values (list): an array of values for partitions

        Returns: None

        """
        partition_kv = ["{}='{}'".format(key, value) for key, value in
                        zip(partition_keys, partition_values)]
        partition_query = """
        ALTER TABLE {table_name} ADD IF NOT EXISTS PARTITION ({partitions});
        """.format(table_name=table,
                   partitions=', '.join(partition_kv))
        self.run_query(query_string=partition_query)

    def _build_create_table_sql(self, table_settings):
        exists = _construct_table_exists_ddl(table_settings["exists"])
        partitions = _construct_table_partition_ddl(table_settings["partitions"])
        table_properties = _construct_table_properties_ddl(
            table_settings.get("skip_headers", False),
            table_settings["storage_format_selector"].lower(),
            table_settings["encryption"])

        sql = """
            CREATE EXTERNAL TABLE {exists} {table}(
              {columns}
              )
            {partitions}
            ROW FORMAT SERDE 
              '{row_format_serde}' 
            STORED AS INPUTFORMAT 
              '{inputformat}' 
            OUTPUTFORMAT 
              '{outputformat}'
            LOCATION
              's3://{s3_bucket}/{s3_dir}'
            {table_properties}
            """.format(table=table_settings["table"],
                       exists=exists,
                       columns=zip_columns(table_settings["columns"]),
                       partitions=partitions,
                       row_format_serde=self.storage_format_lookup[
                           table_settings["storage_format_selector"]]["row_format_serde"],
                       inputformat=self.storage_format_lookup[
                           table_settings["storage_format_selector"]]["inputformat"],
                       outputformat=self.storage_format_lookup[
                           table_settings["storage_format_selector"]]["outputformat"],
                       s3_bucket=table_settings["s3_bucket"],
                       s3_dir=table_settings["s3_dir"],
                       table_properties=table_properties)
        LOG.info("Query from table settings:\n %s", sql)
        return sql

    def create_table(self, table_settings):
        """
        Create a table from given settings
        Args:
            table_settings (dict): Dictionary of settings to create table

        Returns: None
        """
        table_sql = self._build_create_table_sql(table_settings)
        LOG.info(table_sql)
        self.run_query(table_sql)

    def get_table_ddl(self, table):
        """
        Retrive the table DDL in string
        Args:
            table (string): name of the table for which ddl needs to be generated

        Returns: string containing the athena table DDL

        """
        # Read the ddl of temporary table
        ddl_result = self.run_query("""SHOW CREATE TABLE {}""".format(table), return_result=True)
        ddl = ""
        for row in ddl_result["ResultSet"]["Rows"]:
            for column in row["Data"]:
                ddl = ddl + " " + column["VarCharValue"]
            ddl = ddl + "\n"
        return ddl

    def get_table_data_location(self, table: str) -> tuple:
        """
        Retrieves the table's S3 data location using glue meta store
        Args:
            table (str): name of the table
        Returns: tuple of s3 bucket and key
        """
        table = self.get_glue_table_metadata(table)
        location = table['Table']['StorageDescriptor']['Location']
        bucket = location.split("/")[2]
        key = "/".join(location.split("/")[3:])
        return (bucket, key)

    def get_table_columns(self, table: str) -> Tuple[List[dict], List[dict]]:
        """
        Retrieves the table's columns using glue meta store
        Args:
            table (str): name of the table
        Returns: tuple of regular and partition column dictonaries
        """
        table = self.get_glue_table_metadata(table)
        regular_columns = table["Table"]["StorageDescriptor"]["Columns"]
        partition_columns = table["Table"]["PartitionKeys"]
        return (regular_columns, partition_columns)

    def get_glue_table_metadata(self, table: str) -> dict:
        """
        Get table metadata from glue
        Args:
            table (str): Athena table name
        Returns: A dict of table metadata
        """
        return self.conn.client('glue').get_table(DatabaseName=self.database, Name=table)

    def drop_table(self, table_name):
        """
        Drop a given athena table

        Args:
            table_name (string): name of the table to be dropped

        Returns: None

        """
        self.run_query("""DROP TABLE IF EXISTS {}""".format(table_name))


def generate_csv_ctas(select_query, destination_table, destination_bucket, destination_key,
                      partition_fields=''):
    """
    Method to generate a CTAS query string for creating csv output

    Args:
        select_query (string): the query to be used for table generation
        destination_table (string): name of the new table being created
        destination_bucket (string): the s3 bucket where the data from select query will be stored
        destination_key (string): the s3 directory where the data from select query will be stored
        partition_fields (string): partition field names

    Returns (string): CTAS Query in a string

    """
    return _get_ctas_statement(destination_bucket, destination_key, destination_table,
                               partition_fields, select_query, file_format='TEXTFILE')


def generate_parquet_ctas(select_query, destination_table, destination_bucket, destination_key,
                          partition_fields=''):
    """
    Method to generate a CTAS query string for creating parquet output

    Args:
        select_query (string): the query to be used for table generation
        destination_table (string): name of the new table being created
        destination_bucket (string): the s3 bucket where the data from select query will be stored
        destination_key (string): the s3 directory where the data from select query will be stored
        partition_fields (string): partition field names

    Returns (string): CTAS Query in a string

    """
    return _get_ctas_statement(destination_bucket, destination_key, destination_table,
                               partition_fields, select_query, file_format='parquet')


def _get_ctas_statement(destination_bucket, destination_key, destination_table, partition_fields,
                        select_query, file_format):
    delimiter = ''
    if file_format == 'TEXTFILE':
        delimiter = "field_delimiter=',',"
    partitioned_by = ""
    if partition_fields != '':
        partitioned_by = """,
        partitioned_by = ARRAY[{partition_keys}]
        """.format(partition_keys=partition_fields)
    final_query = """
        CREATE TABLE {destination_table}
        WITH (
            {delimiter}
            format='{file_format}',
            external_location='s3://{bucket}/{key}'{partitioned_by}
        ) AS
        {athena_query}
        """.format(destination_table=destination_table,
                   bucket=destination_bucket,
                   key=destination_key,
                   athena_query=select_query,
                   partitioned_by=partitioned_by,
                   file_format=file_format,
                   delimiter=delimiter)
    return final_query


def zip_columns(column_list):
    """
    Combine the column list into a zipped comma separated list of column name and data type
    Args:
        column_list (list): an array of dictionaries with keys column and type

    Returns (string): a string containing comma separated list of column name and data type

    """
    return ", ".join(["{} {}".format(col['column'], col["type"]) for col in column_list])


def _construct_table_partition_ddl(partitions):
    partition_query = ""
    if partitions:
        partition_query = """
        PARTITIONED BY ( 
          {columns}
          )
          """.format(columns=zip_columns(partitions))
    return partition_query


def _construct_table_exists_ddl(enable_exists):
    exists = ""
    if enable_exists:
        exists = "IF NOT EXISTS"
    return exists


def _construct_table_properties_ddl(skip_headers, storage_format_selector, encryption):
    if storage_format_selector == "csv" and skip_headers:
        no_of_skip_lines = 1
        table_properties = """
            TBLPROPERTIES ('has_encrypted_data'='{encryption}', 
            'skip.header.line.count'='{no_of_lines}')
            """.format(encryption=str(encryption).lower(),
                       no_of_lines=no_of_skip_lines)
    else:
        table_properties = """
            TBLPROPERTIES ('has_encrypted_data'='{encryption}')
            """.format(encryption=str(encryption).lower())
    return table_properties


def _get_data_frame_column_types(data_frame):
    data_frame_col_dict = {}
    for col in data_frame:
        data_frame_col_dict[col] = type(data_frame[col][0]).__name__
    return data_frame_col_dict


def get_athena_columns_from_dataframe(data_frame: DataFrame) -> List[dict]:
    """
    Extracts a dictionary of column names and their athena data types from the dataframe
    Args:
        data_frame (DataFrame): the dataframe which the columns need to be extracted
    Returns: list of dict
    """
    column_dtype = _get_data_frame_column_types(data_frame)
    return [
        {"column": field_name, "type": _PYTHON_TO_ATHENA_DATA_TYPE_MAP.get(field_type, "STRING")}
        for
        field_name, field_type in column_dtype.items()]



"""
Setup and release this package for wider consumption using setup tools
"""
import os

import sys
from setuptools import setup, find_packages
from setuptools.command.test import test as test_command


def get_release_version():
    """
    Gets the Release version based on the latest git tag from GIT_TAG env var, else returns 0.0
    Returns: string containing version for the release

    """
    git_version = os.getenv("GIT_TAG", "v0.0")
    pypi_version = git_version.lstrip("v").strip()
    return pypi_version


def get_long_description():
    """
    Get the contents of readme file as long_description
    Returns: bytes containing readme file

    """
    file_path = os.path.abspath(os.path.join(os.path.dirname(__file__), 'README.md'))
    with open(file_path) as readme_file:
        return readme_file.read()


class PyTest(test_command):
    """
    Setup tools class used to initiate pytests
    """
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        """
        Added options to apply to the pytests
        Returns: None
        """
        test_command.initialize_options(self)
        self.pytest_args = ["-vv"]

    def finalize_options(self):
        """
        Options for finalisation
        Returns: None
        """
        test_command.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        """Test executions options"""
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


setup(
    long_description=get_long_description(),
    long_description_content_type='text/markdown',
    packages=find_packages(include=["hip_data_tools", "hip_data_tools.*"]),
    zip_safe=False,
    entry_points={'console_scripts': [
        'version-tracker=hip_data_tools.hipages.version_tracking:main']},
    install_requires=[
        "stringcase",
        "attrs",
        "boto3",
        "joblib",
        "pandas",
        "GitPython",
        "confluent-kafka",
        "pyarrow",
        "cassandra-driver",
        "tqdm",
        "retrying",
        "arrow",
        "s3fs",
        "dataclasses",
        "oauth2client",
        "gspread",
        "googleads",
    ],
    test_suite="tests",
    tests_require=[
        'pytest',
        'pytest-mock==1.10.1',
        "moto",
        "pyarrow",
        "python-snappy",
        "pytest-stub",
        "freezegun==0.1.11",
        "testcontainers",
    ],
    cmdclass={'test': PyTest},
    python_requires='~=3.6',
    version=get_release_version(),

)



"""
handle ETL of data from Athena to Cassandra
"""
from attr import dataclass

from hip_data_tools.apache.cassandra import CassandraConnectionSettings
from hip_data_tools.aws.athena import AthenaUtil
from hip_data_tools.aws.common import AwsConnectionSettings, AwsConnectionManager
from hip_data_tools.etl.s3_to_cassandra import S3ToCassandraSettings, S3ToCassandra


@dataclass
class AthenaToCassandraSettings:
    """S3 to Cassandra ETL settings"""
    source_database: str
    source_table: str
    source_connection_settings: AwsConnectionSettings
    destination_keyspace: str
    destination_table: str
    destination_table_primary_keys: list
    destination_connection_settings: CassandraConnectionSettings
    destination_table_options_statement: str = ""
    destination_batch_size: int = 1


class AthenaToCassandra(S3ToCassandra):
    """
    Class to transfer parquet data from s3 to Cassandra
    Args:
        settings (AthenaToCassandraSettings): the settings around the etl to be executed
    """

    def __init__(self, settings: AthenaToCassandraSettings):
        self.__settings = settings
        self._athena = AthenaUtil(
            database=self.__settings.source_database,
            conn=AwsConnectionManager(self.__settings.source_connection_settings))
        (bucket, key) = self._athena.get_table_data_location(self.__settings.source_table)
        super().__init__(S3ToCassandraSettings(
            source_bucket=bucket,
            source_key_prefix=key,
            source_connection_settings=self.__settings.source_connection_settings,
            destination_keyspace=self.__settings.destination_keyspace,
            destination_table=self.__settings.destination_table,
            destination_table_primary_keys=self.__settings.destination_table_primary_keys,
            destination_table_options_statement=self.__settings.destination_table_options_statement,
            destination_batch_size=self.__settings.destination_batch_size,
            destination_connection_settings=self.__settings.destination_connection_settings,
        ))



"""
Module to deal with data transfer from S3 to Cassandra
"""
from typing import Optional, List

from attr import dataclass

from hip_data_tools.aws.common import AwsConnectionSettings, AwsConnectionManager
from hip_data_tools.aws.s3 import S3Util
from hip_data_tools.common import LOG


@dataclass
class S3ToS3Settings:
    """S3 to Cassandra ETL settings"""
    source_bucket: str
    source_key_prefix: str
    suffix: Optional[str]
    target_bucket: str
    target_key_prefix: str
    connection_settings: AwsConnectionSettings


class S3ToS3:
    """
    Class to transfer objects from s3 to s3
    Args:
        settings (S3ToS3Settings): settings for the etl to be executed
    Eg:
    >>> etl = S3ToS3(
    ...        S3ToS3Settings(
    ...            source_bucket=source_bucket,
    ...            source_key_prefix="source/prefix",
    ...            suffix=None,
    ...            target_bucket=target_bucket,
    ...            target_key_prefix="target/prefix",
    ...            connection_settings=aws_setting
    ...        )
    ...    )
    """

    def __init__(self, settings: S3ToS3Settings):
        self.__settings = settings
        self._s3_util = None
        self._source_keys = None

    def _get_s3_util(self) -> S3Util:
        if self._s3_util is None:
            self._s3_util = S3Util(
                bucket=self.__settings.source_bucket,
                conn=AwsConnectionManager(self.__settings.connection_settings),
            )
        return self._s3_util

    def _get_target_key(self, source_key: str) -> str:
        file_name = source_key.split('/')[-1]
        return f"{self.__settings.target_key_prefix}/{file_name}"

    def list_source_files(self) -> List[str]:
        """
        List all eligible source keys
        Returns: List[str]
        """
        if self._source_keys is None:
            keys = self._get_s3_util().get_keys(self.__settings.source_key_prefix)
            if self.__settings.suffix:
                keys = [key for key in keys if key.endswith(self.__settings.suffix)]
            self._source_keys = keys
            LOG.info("Listed and cached %s source files", len(self._source_keys))
        return self._source_keys

    def transfer_file(self, source_key: str) -> None:
        """
        Transfer one source s3 key to target
        Args:
            source_key (str): source s3 key
        Returns: None
        """
        s3 = self._get_s3_util().get_client()
        copy_source = {
            'Bucket': self.__settings.source_bucket,
            'Key': source_key
        }
        target_key = self._get_target_key(source_key)
        LOG.info("Transferring Key s3://%s/%s to s3://%s/%s",
                 self.__settings.source_bucket,
                 source_key,
                 self.__settings.target_bucket,
                 target_key)
        s3.copy(copy_source, self.__settings.target_bucket, target_key)

    def transfer_all_files(self) -> None:
        """
        Transfer all source keys to target, sequentially
        Returns: None
        """
        for source_key in self.list_source_files():
            self.transfer_file(source_key)



from abc import ABC
from typing import Any, Optional

import boto3 as boto
from attr import dataclass
from botocore.client import BaseClient

from hip_data_tools.common import KeyValueSource, ENVIRONMENT, SecretsManager


class AwsSecretsManager(SecretsManager):
    """

    Args:
        source (KeyValueSource): a kv source that has secrets
        access_key_id_var (str): the variable name or the key for finding access_key_id
        secret_access_key_var (str): the variable name or the key for finding secret_access_key
        use_session_token (bool): flag to check the session token is required or not
        aws_session_token_var (str): the variable name or the key for finding aws_session_token
    """

    def __init__(self,
                 source: KeyValueSource = ENVIRONMENT,
                 access_key_id_var: str = "AWS_ACCESS_KEY_ID",
                 secret_access_key_var: str = "AWS_SECRET_ACCESS_KEY",
                 use_session_token: bool = False,
                 aws_session_token_var: str = "AWS_SESSION_TOKEN"):
        self._required_keys = [access_key_id_var, secret_access_key_var, ]
        if use_session_token:
            self._required_keys.append(aws_session_token_var)
        super().__init__(self._required_keys, source)

        self.aws_access_key_id = self.get_secret(access_key_id_var)
        self.aws_secret_access_key = self.get_secret(secret_access_key_var)
        self.aws_session_token = self.get_secret(aws_session_token_var)


@dataclass
class AwsConnectionSettings:
    """Encapsulates the Cassandra connection settings"""
    region: str
    profile: Optional[str]
    secrets_manager: Optional[AwsSecretsManager]


class AwsConnectionManager:
    """
    utility class to connect to a database and perform some basic operations
    example -
    to connect using an aws cli profile
    >>> conn = AwsConnectionManager(
    ...     AwsConnectionSettings(region_name="ap-southeast-2", profile="default",
    secrets_manager=None))

    # OR if you want to connect using the standard aws environment variables
    (aws_access_key_id, aws_secret_access_key):
    >>> conn = AwsConnectionManager(settings=AwsConnectionSettings(region_name="ap-southeast-2",
    profile=None, secrets_manager=AwsSecretsManager()))

    # OR if you want custom set of env vars to connect
    >>> conn = AwsConnectionManager(
    ...     settings=AwsConnectionSettings(
    ...         region_name="ap-southeast-2",
    ...         secrets_manager=AwsSecretsManager(
    ...             access_key_id_var="SOME_CUSTOM_AWS_ACCESS_KEY_ID",
    ...             secret_access_key_var="SOME_CUSTOM_AWS_SECRET_ACCESS_KEY",
    ...             use_session_token=True,
    ...             aws_session_token_var="SOME_CUSTOM_AWS_SESSION_TOKEN"
    ...             ),
    ...         profile=None,
    ...         )
    ...     )

    Args:
        settings (AwsConnectionSettings): settings to use for connecting to aws
    """

    def __init__(self, settings: AwsConnectionSettings):
        self.settings = settings
        self._session = None

    def client(self, client_type):
        """
        Get a client for specific aws service
        Args:
            client_type (string): choice of aws service like s3, athena, etc. based on boto3:
            session.client(...)

        Returns (client): boto3 client

        """
        return self._get_session().client(client_type, region_name=self.settings.region)

    def resource(self, resource_type):
        """
        Get a resource for specific aws service
        Args:
            resource_type (string): choice of aws service like s3, athena, etc. based on boto3:
            session.client(...)
        Returns (resource): boto3 of type resource_type
        """
        return self._get_session().resource(resource_type, region_name=self.settings.region)

    def _get_session(self):
        """
        Connect and provide an aws session object
        Returns: Session object
        """
        if self._session is None:
            if self.settings.profile is not None:
                self._session = boto.Session(profile_name=self.settings.profile)
            else:
                self._session = boto.Session(
                    aws_access_key_id=self.settings.secrets_manager.aws_access_key_id,
                    aws_secret_access_key=self.settings.secrets_manager.aws_secret_access_key,
                    aws_session_token=self.settings.secrets_manager.aws_session_token,
                )
        return self._session


class AwsUtil(ABC):
    """
    Common Aws class to use boto connection and resources
    Args:
        conn (AwsConnectionManager): Connection to use for accessing aws resources
        boto_type: the type of boto client / resource to instantiate
    """

    def __init__(self, conn: AwsConnectionManager, boto_type: str):
        self.conn = conn
        self._client = None
        self._resource = None
        self.boto_type = boto_type

    def get_client(self) -> BaseClient:
        """
        returns a boto client and creates one if not present
        Returns: BaseClient
        """
        if self._client is None:
            self._client = self.conn.client(self.boto_type)
        return self._client

    def get_resource(self) -> Any:
        """
        returns a boto resporce and creates one if not present
        Returns: Any
        """
        if self._resource is None:
            self._resource = self.conn.resource(self.boto_type)
        return self._resource



"""
Module contains variables and methods used for common / shared operations throughput the package
"""
import json
import logging
import os
import re
import uuid
from abc import ABC, abstractmethod
from collections import OrderedDict
from typing import List

import stringcase
from pandas import DataFrame

COMMON_INTEGER_FIELDS = ["id", "campaign_id", "base_ad_group_id"]

LOG = logging.getLogger(__name__)
"""
logger object to handle logging in the entire package
"""


def _generate_random_file_name():
    random_tmp_file_nm = "/tmp/tmp_file{}".format(str(uuid.uuid4()))
    return random_tmp_file_nm


def get_from_env_or_default_with_warning(env_var, default_val):
    """
    Get environmental variables or, if they aren't present, default to a
    specific value

    Args:
        env_var (str): Name of the environmental variable to read
        default_val (any): Value to default to if relevant env var is not
                        present
    Returns (Any): Value
    """
    value = os.environ.get(env_var)
    if value is None:
        warning_string = "Environmental variable {} not found, " \
                         "defaulting to {}".format(env_var,
                                                   str(default_val))
        LOG.warning(warning_string)
        value = default_val
    return value


class KeyValueSource(ABC):
    """
    Abstract class for sourcing secrets, it is a key value source for retrieving values for keys
    """

    @abstractmethod
    def get(self, key: str) -> str:
        """
        Abstract method to get the value for a given key

        Args:
            key (str): the key for which the value needs to be returned
        Returns: str
        """

    @abstractmethod
    def exists(self, key):
        """
        Abstract methosd to verify if a key exists in the given data store
        Args:
            key (str): the key to be verified for existance
        Returns: bool
        """


class EnvironmentKeyValueSource(KeyValueSource):
    """
    class for sourcing secrets from env variables
    """

    def exists(self, key):
        """
        verify if a key exists

        Args:
            key (str): the key to be verified for existance
        Returns: bool
        """
        if os.getenv(key) is None:
            return False
        return True

    def get(self, key):
        """
        get the value for a given key

        Args:
            key (str): the key for which the value needs to be returned
        Returns: str
        """
        return os.getenv(key)


class DictKeyValueSource(KeyValueSource):
    """
    class for sourcing secrets from a provided Dict object, usually used for testing
    """

    def __init__(self, data):
        self.data = data

    def exists(self, key):
        """
        verify if a key exists

        Args:
            key (str): the key to be verified for existance
        Returns: bool
        """
        if key in self.data:
            return True
        return False

    def get(self, key):
        """
        get the value for a given key

        Args:
            key (str): the key for which the value needs to be returned
        Returns: str
        """
        return self.data[key]


ENVIRONMENT: EnvironmentKeyValueSource = EnvironmentKeyValueSource()
"""
Standard Environment Variable Secret source to be reused across the project
"""


class SecretsManager(ABC):
    """
    A secret management abstract class that provides ways of extracting secrets
    The class allows a subsequent connection class to use env vars to extract secrets in a
    structured manner

    Args:
        required_keys (list[str]): a list of keys which will be checked for existence
        source (KeyValueSource): a kv source that has secrets
    """

    def __init__(self, required_keys: list, source: KeyValueSource):
        self.keys = required_keys
        self._source = source
        for key in self.keys:
            if not self._source.exists(key):
                raise Exception("Required Environment Variable {} does not exist!".format(key))

    def get_secret(self, key):
        """
        get the secret valye for a given key

        Args:
            key (str): the key for given secret
        Returns: str
        """
        return self._source.get(key)


def flatten_nested_dict(data: dict, delimiter: str = "_", snake_cased_keys: bool = True) -> dict:
    """
    takes arbitrarily nested levels of a dictionary and un nests it to one level of key value pairs

    Args:
        data (dict): the dictionary ro be un nested
        delimiter (str): the delimiter to concatenate nested keys
        snake_cased_keys (bool): convert the keys in resulting dict to snake case
    Returns: dict
    """

    def expand(key, value):
        if isinstance(value, dict) or isinstance(value, OrderedDict):
            return [(key + delimiter + k, v) for k, v in flatten_nested_dict(value).items()]
        else:
            return [(key, value)]

    items = [item for k, v in data.items() for item in expand(k, v)]
    if snake_cased_keys:
        items = [(to_snake_case(k), v) for k, v in items]
    return dict(items)


camel_case_detect = re.compile(r'(?<!^)(?=[A-Z])')
"""Regex pattern to be reused for Camel Case"""

special_characters_detect = re.compile(r'[^a-zA-Z0-9]')
"""Regex pattern to detect special characters"""


def to_snake_case(column_name: str) -> str:
    """
    Converts the column name to Athena compatible snake_case

    Args:
        column_name (str): column name string to be sanitized
    Returns: str
    """
    # Detect and replace special_chars
    str_replaced_special_chars = special_characters_detect.sub('_', column_name)
    # Convert to Snake Case
    camel_column_name = stringcase.snakecase(str_replaced_special_chars)
    return camel_column_name


def nested_list_of_dict_to_dataframe(data: List[dict]) -> DataFrame:
    flattened_dicts = [flatten_nested_dict(d) for d in data]
    df = DataFrame(data=flattened_dicts)
    for common_int_field in COMMON_INTEGER_FIELDS:
        if common_int_field in df.columns:
            df[common_int_field].fillna(0, inplace=True)
            df[common_int_field] = df[common_int_field].astype(int)
    for col in list(df):
        _convert_complex_val_to_json(col, df)
    return df


def _convert_complex_val_to_json(col, df):
    col_type = type(next(iter(df[col].values), object))
    if col_type is list:
        try:
            df[col] = df[col].apply(lambda x: json.dumps(x))
        except ValueError:
            df[col] = df[col].astype(str)


def dataframe_columns_to_snake_case(data: DataFrame) -> None:
    original_columns = data.columns
    data.columns = [to_snake_case(col) for col in original_columns]



"""
handle ETL of data from Athena to Cassandra
"""
from attr import dataclass

from hip_data_tools.aws.athena import AthenaUtil
from hip_data_tools.aws.common import AwsConnectionSettings, AwsConnectionManager
from hip_data_tools.etl.s3_to_dataframe import S3ToDataFrameSettings, S3ToDataFrame


@dataclass
class AthenaToDataFrameSettings:
    """S3 to Cassandra ETL settings"""
    source_database: str
    source_table: str
    source_connection_settings: AwsConnectionSettings


class AthenaToDataFrame(S3ToDataFrame):
    """
    Class to transfer parquet data from s3 to Cassandra
    Args:
        settings (AthenaToCassandraSettings): the settings around the etl to be executed
    """

    def __init__(self, settings: AthenaToDataFrameSettings):
        self._athena = None
        self.__settings = settings
        (bucket, key) = self._get_athena_util().get_table_data_location(settings.source_table)
        self.s3_settings = S3ToDataFrameSettings(
            source_bucket=bucket,
            source_key_prefix=key,
            source_connection_settings=settings.source_connection_settings,
        )
        super().__init__(self.s3_settings)

    def _get_athena_util(self):
        if self._athena is None:
            self._athena = AthenaUtil(
                database=self.__settings.source_database,
                conn=AwsConnectionManager(self.__settings.source_connection_settings))
        return self._athena
